import logging
import requests
import sys
import json
import esgfpid.utils
import esgfpid.exceptions
import esgfpid.defaults
from esgfpid.utils import loginfo, logdebug, logtrace, logerror, logwarn

LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())

'''
This class is only responsible for sending queries
(that were assembled elsewhere) to a solr instance,
validate and parse the result and return it.

It basically only provides one method,
:func:`~solr.SolrServerConnector.send_query`.

'''
class SolrServerConnector(object):

    '''
    Create an instance.

    :param solr_url: Mandatory.
    :param https_verify: Mandatory. Boolean.
    :param disable_insecure_request_warning: Mandatory. Boolean.
    '''
    def __init__(self, **args):
        self.__check_presence_of_args(args)
        self.__set_attributes(args)
        self.__disable_warning_if_desired(args)

    def __check_presence_of_args(self, args):
        mandatory_args = ['solr_url', 'https_verify', 'disable_insecure_request_warning']
        esgfpid.utils.check_presence_of_mandatory_args(args, mandatory_args)
        esgfpid.utils.check_noneness_of_mandatory_args(args, mandatory_args)

    def __set_attributes(self, args):
        self.__solr_url = args['solr_url'].strip('/')
        self.__https_verify = args['https_verify']
        self.__solr_headers = {'Accept': 'application/json,text/json', 'Content-Type': 'application/json'}

    def __disable_warning_if_desired(self, args):
        if args['disable_insecure_request_warning']:
            self.__disable_insecure_request_warning()

    def __disable_insecure_request_warning(self):
        logdebug(LOGGER, 'Disabling insecure request warnings...')
        # Source: http://stackoverflow.com/questions/27981545/suppress-insecurerequestwarning-unverified-https-request-is-being-made-in-pytho#28002687
        from requests.packages.urllib3.exceptions import InsecureRequestWarning
        requests.packages.urllib3.disable_warnings(InsecureRequestWarning)

    '''
    Send a message to a solr instance.

    This raises SolrError exceptions in various situations:
     - Connection problems
     - Response is empty, None, or invalid JSON
     - Response returns 404 or any other HTTP code other than 200

     :raises: esgfpid.exceptions.SolrError
     :return: A JSON response. It never returns None.
    '''
    def send_query(self, query):
        response = self.__get_request_to_solr(query)
        self.__check_response_for_error_codes(response)
        response_json = self.__get_json_from_response(response)
        return response_json

    def __get_request_to_solr(self, query_dict):
        logdebug(LOGGER, 'Sending GET request to solr at "'+self.__solr_url+'".')
        try:
            resp = requests.get(
                self.__solr_url,
                params=query_dict,
                headers=self.__solr_headers,
                verify=self.__https_verify
            )
            return resp
        except requests.exceptions.ConnectionError as e:
            msg = 'Could not connect to solr: ConnectionError.'
            LOGGER.exception(msg)
            raise esgfpid.exceptions.SolrError(msg)

    def __get_json_from_response(self, response):
        try:
            response_json = json.loads(response.content)
            logdebug(LOGGER, 'Solr response ok, returned JSON content.')
            return response_json

        except (ValueError, TypeError) as e:
            msg = 'Error while parsing Solr response. It seems to be no valid JSON. Message: '+repr(e)
            logerror(LOGGER, msg)
            raise esgfpid.exceptions.SolrError(msg)

    def __check_response_for_error_codes(self, response):

        if response is None:
            msg = 'Solr returned no response (None)'
            logerror(LOGGER, msg)
            raise esgfpid.exceptions.SolrError(msg)

        elif response.status_code == 200:
            if response.content is None:
                msg = 'Solr returned an empty response (with content None)'
                logerror(LOGGER, msg)
                raise esgfpid.exceptions.SolrError(msg)

        elif response.status_code == 404:
            msg = 'Solr returned no response (HTTP 404)'
            logerror(LOGGER, msg)
            raise esgfpid.exceptions.SolrError(msg)

        else:
            msg = 'Solr replied with code '+str(response.status_code)
            logerror(LOGGER, msg)
            raise esgfpid.exceptions.SolrError(msg)





import unittest
import mock
import logging
import json
import requests
import esgfpid.solr.solr
import tests.resources.responsemock


# Logging:
LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())

# Load some data that is needed for testing
PATH_RES = tests.utils.get_super_neighbour_directory(__file__, 'resources')
SOLR_RESPONSE = json.load(open(PATH_RES+'/solr_response.json'))

# Test resources:
import resources.TESTVALUES as TESTHELPERS

'''
Unit tests for module esgfpid.solr.serverconnector.

This module has no references to other modules, so
no other modules need to be used or mocked.

However, it talks to a solr server (via requests
module), so we need to mock that one. We patch all
calls to requests.get()
'''
class SolrServerConnectorTestCase(unittest.TestCase):

    def setUp(self):
        LOGGER.info('######## Next test (%s) ##########', __name__)

    def tearDown(self):
        LOGGER.info('#############################')

    def test_init_ok(self):

        # Run code to be tested:
        testsolr = esgfpid.solr.serverconnector.SolrServerConnector(
            solr_url = 'blablabla///',
            https_verify = False,
            disable_insecure_request_warning = False
        )

        # Check result
        self.assertIsInstance(testsolr, esgfpid.solr.serverconnector.SolrServerConnector, 'Constructor fail.')
        self.assertFalse(testsolr._SolrServerConnector__https_verify, 'HTTPS verify not set to False.')
        self.assertEqual(testsolr._SolrServerConnector__solr_url, 'blablabla', 'Wrong url.')


    def test_init_with_optional_args(self):

        # Run code to be tested:
        testsolr = esgfpid.solr.serverconnector.SolrServerConnector(
            solr_url = 'blablabla',
            https_verify = True,
            disable_insecure_request_warning = True
        )

        self.assertIsInstance(testsolr, esgfpid.solr.serverconnector.SolrServerConnector, 'Constructor fail.')
        self.assertTrue(testsolr._SolrServerConnector__https_verify, 'HTTPS verify not set to True.')

    @mock.patch('esgfpid.solr.serverconnector.requests.get')
    def test_send_query_ok_patched(self, getpatch):

        # Define the replacement for the patched method:
        mock_response = tests.resources.responsemock.MockSolrResponse(success=True)
        getpatch.return_value = mock_response

        # Test variables
        testsolr = TESTHELPERS.get_testsolr_connector()

        # Run code to be tested:
        response = testsolr.send_query('blah')

        # Check result:
        # We expect the mock response to be returned and parsed to JSON.
        self.assertTrue('responseHeader' in response, 'Solr response was not correct JSON: '+str(response))
        self.assertTrue(len(response['facet_counts']['facet_fields']['bla'])==4, 'JSON content was not transmitted correctly.')

    @mock.patch('esgfpid.solr.serverconnector.requests.get')
    def test_send_query_200_empty_patched(self, getpatch):

        # Define the replacement for the patched method:
        mock_response = tests.resources.responsemock.MockSolrResponse(success=True, content='NONE')
        getpatch.return_value = mock_response

        # Test variables
        testsolr = TESTHELPERS.get_testsolr_connector()

        # Run code to be tested and check exception:
        with self.assertRaises(esgfpid.exceptions.SolrError) as raised:
            response = testsolr.send_query('blah')
        self.assertIn('empty response', repr(raised.exception))

    @mock.patch('esgfpid.solr.serverconnector.requests.get')
    def test_send_query_broken_json_patched(self, getpatch):

        # Define the replacement for the patched method:
        broken_json = '{"responseHeader":{}, "response":{}, "facet_counts": {}'
        mock_response = tests.resources.responsemock.MockSolrResponse(success=True, content=broken_json)
        getpatch.return_value = mock_response

        # Test variables
        testsolr = TESTHELPERS.get_testsolr_connector()

        # Run code to be tested and check exception:
        with self.assertRaises(esgfpid.exceptions.SolrError) as raised:
            response = testsolr.send_query('blah')
        self.assertIn('no valid JSON', repr(raised.exception))

    @mock.patch('esgfpid.solr.serverconnector.requests.get')
    def test_send_query_return_none_patched(self, getpatch):

        # Define the replacement for the patched method:
        getpatch.return_value = None

        # Test variables
        testsolr = TESTHELPERS.get_testsolr_connector()

        # Run code to be tested and check exception:
        with self.assertRaises(esgfpid.exceptions.SolrError) as raised:
            response = testsolr.send_query('blah')
        self.assertIn('no response', repr(raised.exception))

    @mock.patch('esgfpid.solr.serverconnector.requests.get')
    def test_send_query_http_404_patched(self, getpatch):

        # Define the replacement for the patched method:
        mock_response = tests.resources.responsemock.MockSolrResponse(notfound=True)
        getpatch.return_value = mock_response

        # Test variables
        testsolr = TESTHELPERS.get_testsolr_connector()

        # Run code to be tested and check exception:
        with self.assertRaises(esgfpid.exceptions.SolrError) as raised:
            response = testsolr.send_query('blah')
        self.assertIn('HTTP 404', repr(raised.exception))

    @mock.patch('esgfpid.solr.serverconnector.requests.get')
    def test_send_query_patched_connection_error(self, getpatch):

        # Define the replacement for the patched method:
        getpatch.side_effect = requests.exceptions.ConnectionError()

        # Test variables
        testsolr = TESTHELPERS.get_testsolr_connector()

        # Run code to be tested and check exception:
        with self.assertRaises(esgfpid.exceptions.SolrError) as raised:
            response = testsolr.send_query('blah')
        self.assertIn('ConnectionError', repr(raised.exception))

    @mock.patch('esgfpid.solr.serverconnector.requests.get')
    def test_send_query_patched_5000(self, getpatch):

        # Define the replacement for the patched method:
        mock_response = tests.resources.responsemock.MockSolrResponse(status_code=5000)
        getpatch.return_value = mock_response

        # Test variables
        testsolr = TESTHELPERS.get_testsolr_connector()

        # Run code to be tested and check exception:
        with self.assertRaises(esgfpid.exceptions.SolrError) as raised:
            response = testsolr.send_query('blah')
        self.assertIn('code 5000', repr(raised.exception))



import logging
import esgfpid.utils
from esgfpid.utils import loginfo, logdebug, logtrace, logerror, logwarn


LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())
        
class AssistantOneVersion(object):
 
    def __init__(self, **args):
        mandatory_args = ['drs_id', 'data_node', 'prefix', 'coupler']
        esgfpid.utils.check_presence_of_mandatory_args(args, mandatory_args)

        self._coupler = args['coupler']
        self._drs_id = args['drs_id']
        self._data_node = args['data_node'].rstrip('/')
        self._prefix = args['prefix']

    def unpublish_one_dataset_version(self, **args):
        optional_args = ['dataset_handle', 'version_number']
        esgfpid.utils.add_missing_optional_args_with_value_none(args, optional_args)

        handle = args['dataset_handle']
        version_number = args['version_number']

        if handle and version_number:
            self.__both_given(handle, version_number)
            loginfo(LOGGER, 'Requesting to unpublish version %s of dataset %s from %s (handle: %s).', version_number, self._drs_id, self._data_node, handle)
        elif handle:
            self.__only_handle_given(handle)
            loginfo(LOGGER, 'Requesting to unpublish a version of dataset %s from %s (handle: %s).', self._drs_id, self._data_node, handle)
        elif version_number:
            self.__only_version_given(version_number)
            loginfo(LOGGER, 'Requesting to unpublish version %s of dataset %s from %s.', version_number, self._drs_id, self._data_node)
        else:
            msg = 'Neither a handle nor a version number were specified for unpublication!'
            raise esgfpid.exceptions.ArgumentError(msg)

    def __both_given(self, handle, version_number):
        self.__check_whether_handle_and_version_number_match(handle, version_number)
        logdebug(LOGGER, 'Unpublishing dataset with handle %s (one version, %s).', handle, version_number)
        message = self.__make_message(handle, version_number=version_number)
        self._send_message_to_queue(message)

    def __only_handle_given(self, handle):
        logdebug(LOGGER, 'Unpublishing dataset with handle %s (one version).', handle)
        message = self.__make_message(handle)
        self._send_message_to_queue(message)

    def __only_version_given(self, version_number):
        handle = self.__derive_handle_from_version(version_number)
        logdebug(LOGGER, 'Unpublishing dataset with handle %s (one version, %s). Handle was derived from version number', handle, version_number)
        message = self.__make_message(handle, version_number=version_number)
        self._send_message_to_queue(message)

    def __derive_handle_from_version(self, version_number):
        derived_handle = esgfpid.utils.make_handle_from_drsid_and_versionnumber(
            drs_id=self._drs_id,
            version_number=version_number,
            prefix=self._prefix
        )
        return derived_handle

    def __check_whether_handle_and_version_number_match(self, specified_handle, version_number):
        derived_handle = self.__derive_handle_from_version(version_number)
        if derived_handle != specified_handle:
            msg = ('Given handle not the same as computed handle. '+
                   'Given: '+specified_handle+', '+
                   'computed: '+derived_handle+
                   ' (from '+self._drs_id+' and '+str(version_number)+').'
            )
            LOGGER.error(msg)
            raise ValueError(msg)
            # TODO COMPLAIN LOUDLY

    def __make_message(self, handle, **kwargs):
        message_timestamp = esgfpid.utils.get_now_utc_as_formatted_string()

        message = esgfpid.assistant.messages.unpublish_one_version(
            timestamp = message_timestamp,
            data_node = self._data_node,
            dataset_handle = handle,
            drs_id = self._drs_id # can be none
        )
        if 'version_number' in kwargs:
            message['version_number'] = int(kwargs['version_number'])
        return message

    def _send_message_to_queue(self, message):
        self._coupler.send_message_to_queue(message)

class AssistantAllVersions(AssistantOneVersion):

    def __init__(self, **args):
        super(AssistantAllVersions, self).__init__(**args)
        if 'consumer_solr_url' not in list(args.keys()):
            args['consumer_solr_url'] = None
        self.__consumer_solr_url = args['consumer_solr_url']

    def unpublish_all_dataset_versions(self):

        # If solr is switched off, consumer must find versions:
        if self._coupler.is_solr_switched_off():
            self.__unpublish_allversions_consumer_must_find_versions()

        # Get handles or version numbers from solr:
        else:
            all_handles_or_versionnumbers = self.__get_all_handles_or_versionnumbers()
            all_handles = all_handles_or_versionnumbers['dataset_handles']
            all_version_numbers = all_handles_or_versionnumbers['version_numbers']

            # If we can have all versions' handles, it's easy.
            if all_handles is not None:
                self.__unpublish_all_dataset_versions_by_handle(all_handles)

            # If not, we have the version numbers (and can make the handles from them):
            elif all_version_numbers is not None:
                self.__unpublish_all_dataset_versions_by_version(all_version_numbers)

            # If neither, let the consumer find them
            else:
                self.__unpublish_allversions_consumer_must_find_versions()
                loginfo(LOGGER, 'Requesting to unpublish all versions of dataset %s from %s', self._drs_id, self._data_node)

    def __unpublish_all_dataset_versions_by_version(self, all_version_numbers):
        for version_number in all_version_numbers:
            self.unpublish_one_dataset_version(version_number=version_number)
  
    def __unpublish_all_dataset_versions_by_handle(self, all_handles):
        for handle in all_handles:
            self.unpublish_one_dataset_version(dataset_handle=handle)

    def __unpublish_allversions_consumer_must_find_versions(self):
        message = self.__make_message_case_consumer_must_find_versions()
        self._send_message_to_queue(message)

    def __make_message_case_consumer_must_find_versions(self):
        message_timestamp = esgfpid.utils.get_now_utc_as_formatted_string()

        message = esgfpid.assistant.messages.unpublish_allversions_consumer_must_find_versions(
            timestamp=message_timestamp,
            drs_id=self._drs_id,
            data_node=self._data_node,
            consumer_solr_url=self.__consumer_solr_url)
        return message

    def __get_all_handles_or_versionnumbers(self):
        all_handles_or_versionnumbers = self._coupler.retrieve_datasethandles_or_versionnumbers_of_allversions(
            drs_id=self._drs_id
        )
        return all_handles_or_versionnumbers




import unittest
import logging
import tests.utils as utils
from tests.utils import compare_json_return_errormessage as error_message

from esgfpid.assistant.publish import DatasetPublicationAssistant
from esgfpid.exceptions import SolrSwitchedOff

# Logging
LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())

# Test resources:
from resources.TESTVALUES import *
import resources.TESTVALUES as TESTHELPERS


class PublishTestCase(unittest.TestCase):

    def setUp(self):
        LOGGER.info('######## Next test (%s) ##########', __name__)

    def tearDown(self):
        LOGGER.info('#############################')

    #
    # Testing only init (passing info about dataset, but not about file yet):
    #

    def test_normal_ok(self):

        # Preparations:
        testcoupler = TESTHELPERS.get_coupler()
        TESTHELPERS.patch_solr_returns_empty_file_list(testcoupler)
        args = TESTHELPERS.get_args_for_publication_assistant()

        # Run code to be tested:
        assistant = DatasetPublicationAssistant(coupler=testcoupler, **args)

        # Check result:
        self.assertIsInstance(assistant, DatasetPublicationAssistant,
            'Constructor failed.')

    def test_version_number_is_string_ok(self):

        # Preparations:
        testcoupler = TESTHELPERS.get_coupler(solr_switched_off=True)
        args = TESTHELPERS.get_args_for_publication_assistant()
        args['version_number'] = str(args['version_number'])

        # Run code to be tested:
        assistant = DatasetPublicationAssistant(coupler=testcoupler, **args)

        # Check result:
        self.assertIsInstance(assistant, DatasetPublicationAssistant,
            'Constructor failed.')

    def test_version_number_has_characters_error(self):

        # Preparations:
        testcoupler = TESTHELPERS.get_coupler(solr_switched_off=True)
        args = TESTHELPERS.get_args_for_publication_assistant()
        args['version_number'] = 'abcdef'

        # Run code to be tested and check exception:
        with self.assertRaises(esgfpid.exceptions.ArgumentError):
            assistant = DatasetPublicationAssistant(coupler=testcoupler, **args)

    def test_init_no_consumer_solr_url(self):

        # Preparations:
        testcoupler = TESTHELPERS.get_coupler(solr_switched_off=True)
        args = TESTHELPERS.get_args_for_publication_assistant()
        del args['consumer_solr_url']

        # Run code to be tested and check exception:
        with self.assertRaises(esgfpid.exceptions.ArgumentError):
            assistant = DatasetPublicationAssistant(**args)

    def test_init_consumer_solr_url_is_none(self):

        # Preparations:
        testcoupler = TESTHELPERS.get_coupler(solr_switched_off=True)
        args = TESTHELPERS.get_args_for_publication_assistant()
        args['consumer_solr_url'] = None

        # Run code to be tested:
        assistant = DatasetPublicationAssistant(coupler=testcoupler, **args)

        # Check result:
        self.assertIsInstance(assistant, DatasetPublicationAssistant,
            'Constructor failed.')

    def test_init_string_replica_flag_ok(self):

        # Preparations:
        testcoupler = TESTHELPERS.get_coupler(solr_switched_off=True)
        args = TESTHELPERS.get_args_for_publication_assistant()
        args['is_replica'] = 'False'

        # Run code to be tested:
        assistant = DatasetPublicationAssistant(coupler=testcoupler, **args)

        # Check result:
        self.assertIsInstance(assistant, DatasetPublicationAssistant,
            'Constructor failed.')
        self.assertFalse(assistant._DatasetPublicationAssistant__is_replica,
            'Replica flag should be False if it is passed as "False".')
    
    def test_init_string_wrong_replica_flag(self):

        # Preparations:
        testcoupler = TESTHELPERS.get_coupler(solr_switched_off=True)
        args = TESTHELPERS.get_args_for_publication_assistant()
        args['is_replica'] = 'Maybe'

        # Run code to be tested:
        with self.assertRaises(esgfpid.exceptions.ArgumentError) as raised:
            assistant = DatasetPublicationAssistant(coupler=testcoupler, **args)
        self.assertIn('"Maybe" could not be parsed to boolean', repr(raised.exception),
            'Unexpected error message: %s' % repr(raised.exception))

    #
    # Testing entire publication
    # First publication, no previous info on solr, no consistency check
    #

    def test_normal_publication_ok(self):

        # Preparations:
        testcoupler = TESTHELPERS.get_coupler(solr_switched_off=True)
        TESTHELPERS.patch_with_rabbit_mock(testcoupler)
        dsargs = TESTHELPERS.get_args_for_publication_assistant()
        assistant = DatasetPublicationAssistant(coupler=testcoupler, **dsargs)
        fileargs = TESTHELPERS.get_args_for_adding_file()

        # Run code to be tested:
        assistant.add_file(**fileargs)
        assistant.dataset_publication_finished()

        # Check result (dataset):
        received_rabbit_task = TESTHELPERS.get_received_message_from_rabbitmock(testcoupler, 0)
        expected_rabbit_task = TESTHELPERS.get_rabbit_message_publication_dataset()
        same = utils.is_json_same(expected_rabbit_task, received_rabbit_task)
        self.assertTrue(same, error_message(expected_rabbit_task, received_rabbit_task))

        # Check result (file):
        received_rabbit_task = TESTHELPERS.get_received_message_from_rabbitmock(testcoupler, 1)
        expected_rabbit_task = TESTHELPERS.get_rabbit_message_publication_file()
        same = utils.is_json_same(expected_rabbit_task, received_rabbit_task)
        self.assertTrue(same, error_message(expected_rabbit_task, received_rabbit_task))

    def test_normal_publication_sev_files_ok(self):

        # Test variables
        handle1 = PREFIX_WITH_HDL+'/456'
        handle2 = PREFIX_WITH_HDL+'/789'

        # Preparations:
        testcoupler = TESTHELPERS.get_coupler(solr_switched_off=True)
        TESTHELPERS.patch_with_rabbit_mock(testcoupler)
        dsargs = TESTHELPERS.get_args_for_publication_assistant()
        assistant = DatasetPublicationAssistant(coupler=testcoupler, **dsargs)
        args1 = TESTHELPERS.get_args_for_adding_file()
        args1['file_handle'] = handle1
        args2 = TESTHELPERS.get_args_for_adding_file()
        args2['file_handle'] = handle2

        # Run code to be tested:
        assistant.add_file(**args1)
        assistant.add_file(**args2)
        assistant.dataset_publication_finished()

        # Check result (dataset):
        received_rabbit_task = TESTHELPERS.get_received_message_from_rabbitmock(testcoupler, 0)
        expected_rabbit_task = TESTHELPERS.get_rabbit_message_publication_dataset()
        expected_rabbit_task['files'] = [handle2, handle1]
        same = utils.is_json_same(expected_rabbit_task, received_rabbit_task)
        self.assertTrue(same, error_message(expected_rabbit_task, received_rabbit_task))

        # Check result (file):
        received_rabbit_task = TESTHELPERS.get_received_message_from_rabbitmock(testcoupler, 1)
        expected_rabbit_task = TESTHELPERS.get_rabbit_message_publication_file()
        expected_rabbit_task['handle'] = handle1
        same = utils.is_json_same(expected_rabbit_task, received_rabbit_task)
        self.assertTrue(same, error_message(expected_rabbit_task, received_rabbit_task))

        # Check result (file):
        received_rabbit_task = TESTHELPERS.get_received_message_from_rabbitmock(testcoupler, 2)
        expected_rabbit_task = TESTHELPERS.get_rabbit_message_publication_file()
        expected_rabbit_task['handle'] = handle2
        same = utils.is_json_same(expected_rabbit_task, received_rabbit_task)
        self.assertTrue(same, error_message(expected_rabbit_task, received_rabbit_task))

    def test_add_file_wrong_prefix(self):

        # Preparations:
        testcoupler = TESTHELPERS.get_coupler(solr_switched_off=True)
        dsargs = TESTHELPERS.get_args_for_publication_assistant()
        assistant = DatasetPublicationAssistant(coupler=testcoupler, **dsargs)
        fileargs = TESTHELPERS.get_args_for_adding_file()
        fileargs['file_handle'] = 'hdl:1234/56789'

        # Run code to be tested:
        with self.assertRaises(esgfpid.exceptions.ESGFException) as raised:
            assistant.add_file(**fileargs)
        self.assertIn('ESGF rule violation', repr(raised.exception),
            'Unexpected message: %s' % repr(raised.exception))

    def test_add_file_without_hdl_in_handle(self):

        # Preparations:
        testcoupler = TESTHELPERS.get_coupler(solr_switched_off=True)
        TESTHELPERS.patch_with_rabbit_mock(testcoupler)
        dsargs = TESTHELPERS.get_args_for_publication_assistant()
        assistant = DatasetPublicationAssistant(coupler=testcoupler, **dsargs)
        fileargs = TESTHELPERS.get_args_for_adding_file()
        fileargs['file_handle'] = FILEHANDLE_NO_HDL

         # Run code to be tested:
        assistant.add_file(**fileargs)
        assistant.dataset_publication_finished()

        # Check result (dataset):
        received_rabbit_task = TESTHELPERS.get_received_message_from_rabbitmock(testcoupler, 0)
        expected_rabbit_task = TESTHELPERS.get_rabbit_message_publication_dataset()
        same = utils.is_json_same(expected_rabbit_task, received_rabbit_task)
        self.assertTrue(same, error_message(expected_rabbit_task, received_rabbit_task))

        # Check result (file):
        received_rabbit_task = TESTHELPERS.get_received_message_from_rabbitmock(testcoupler, 1)
        expected_rabbit_task = TESTHELPERS.get_rabbit_message_publication_file()
        same = utils.is_json_same(expected_rabbit_task, received_rabbit_task)
        self.assertTrue(same, error_message(expected_rabbit_task, received_rabbit_task))

    def test_add_file_fileversion_as_integer_filesize_as_string(self):

        # Preparations:
        testcoupler = TESTHELPERS.get_coupler(solr_switched_off=True)
        TESTHELPERS.patch_with_rabbit_mock(testcoupler)
        dsargs = TESTHELPERS.get_args_for_publication_assistant()
        assistant = DatasetPublicationAssistant(coupler=testcoupler, **dsargs)
        fileargs = TESTHELPERS.get_args_for_adding_file()
        fileargs['file_version'] = int(fileargs['file_version'])
        fileargs['file_size'] = str(fileargs['file_size'])

        # Run code to be tested:
        assistant.add_file(**fileargs)
        assistant.dataset_publication_finished()

        # Check result (file):
        received_rabbit_task = TESTHELPERS.get_received_message_from_rabbitmock(testcoupler, 1)
        expected_rabbit_task = TESTHELPERS.get_rabbit_message_publication_file()
        same = utils.is_json_same(expected_rabbit_task, received_rabbit_task)
        self.assertTrue(same, error_message(expected_rabbit_task, received_rabbit_task))

    def test_add_file_filesize_has_characters(self):

        # Preparations:
        testcoupler = TESTHELPERS.get_coupler(solr_switched_off=True)
        dsargs = TESTHELPERS.get_args_for_publication_assistant()
        assistant = DatasetPublicationAssistant(coupler=testcoupler, **dsargs)
        fileargs = TESTHELPERS.get_args_for_adding_file()
        fileargs['file_size'] = 'abc'

        # Run code to be tested:
        with self.assertRaises(esgfpid.exceptions.ArgumentError):
            assistant.add_file(**fileargs)

    def test_normal_publication_without_solr_access(self):
        '''
        Solr was switched off and raises an errors, but this is caught.
        The consistency check is not run, but the rest of the publication is the same.
        '''

        # Preparations:  
        testcoupler = TESTHELPERS.get_coupler()
        TESTHELPERS.patch_with_rabbit_mock(testcoupler)
        TESTHELPERS.patch_solr_raises_error(testcoupler, SolrSwitchedOff) # solr raises error
        dsargs = TESTHELPERS.get_args_for_publication_assistant()
        assistant = DatasetPublicationAssistant(coupler=testcoupler, **dsargs)
        fileargs = TESTHELPERS.get_args_for_adding_file()

        # Run code to be tested:
        assistant.add_file(**fileargs)
        assistant.dataset_publication_finished()

        # Check result (dataset):
        received_rabbit_task = TESTHELPERS.get_received_message_from_rabbitmock(testcoupler, 0)
        expected_rabbit_task = TESTHELPERS.get_rabbit_message_publication_dataset()
        same = utils.is_json_same(expected_rabbit_task, received_rabbit_task)
        self.assertTrue(same, error_message(expected_rabbit_task, received_rabbit_task))

        # Check result (file):
        received_rabbit_task = TESTHELPERS.get_received_message_from_rabbitmock(testcoupler, 1)
        expected_rabbit_task = TESTHELPERS.get_rabbit_message_publication_file()
        same = utils.is_json_same(expected_rabbit_task, received_rabbit_task)
        self.assertTrue(same, error_message(expected_rabbit_task, received_rabbit_task))

    def test_normal_publication_solr_returns_none(self):

        # Preparations:  
        testcoupler = TESTHELPERS.get_coupler()
        TESTHELPERS.patch_with_rabbit_mock(testcoupler)
        TESTHELPERS.patch_solr_returns_previous_files(testcoupler, None) # makes mock return None - this should never happen in reality!
        dsargs = TESTHELPERS.get_args_for_publication_assistant()
        assistant = DatasetPublicationAssistant(coupler=testcoupler, **dsargs)
        fileargs = TESTHELPERS.get_args_for_adding_file()

        # Run code to be tested:
        assistant.add_file(**fileargs)
        assistant.dataset_publication_finished()

        # Check result (dataset):
        received_rabbit_task = TESTHELPERS.get_received_message_from_rabbitmock(testcoupler, 0)
        expected_rabbit_task = TESTHELPERS.get_rabbit_message_publication_dataset()
        same = utils.is_json_same(expected_rabbit_task, received_rabbit_task)
        self.assertTrue(same, error_message(expected_rabbit_task, received_rabbit_task))

    #
    # Testing entire RE-publication
    # Republication, solr provides info on previous files, consistency check
    #

    def test_normal_publication_with_pos_consis_check(self):

        # Test variables:
        fileargs = TESTHELPERS.get_args_for_adding_file()
        prev_list = [FILEHANDLE_HDL]

        # Preparations:
        testcoupler = TESTHELPERS.get_coupler()
        TESTHELPERS.patch_solr_returns_previous_files(testcoupler, prev_list) # solr returns file list
        TESTHELPERS.patch_with_rabbit_mock(testcoupler)
        dsargs = TESTHELPERS.get_args_for_publication_assistant()
        assistant = DatasetPublicationAssistant(coupler=testcoupler, **dsargs)

        # Run code to be tested:
        assistant.add_file(**fileargs)
        assistant.dataset_publication_finished()

        # Check result (dataset):
        received_rabbit_task = TESTHELPERS.get_received_message_from_rabbitmock(testcoupler, 0)
        expected_rabbit_task = TESTHELPERS.get_rabbit_message_publication_dataset()
        same = utils.is_json_same(expected_rabbit_task, received_rabbit_task)
        self.assertTrue(same, error_message(expected_rabbit_task, received_rabbit_task))

        # Check result (file):
        received_rabbit_task = TESTHELPERS.get_received_message_from_rabbitmock(testcoupler, 1)
        expected_rabbit_task = TESTHELPERS.get_rabbit_message_publication_file()
        same = utils.is_json_same(expected_rabbit_task, received_rabbit_task)
        self.assertTrue(same, error_message(expected_rabbit_task, received_rabbit_task))

    def test_normal_publication_with_pos_consis_check_missing_hdl(self):

        # Test variables:
        # File to be published (without hdl):
        fileargs = TESTHELPERS.get_args_for_adding_file()
        fileargs['file_handle'] = FILEHANDLE_NO_HDL
        # Preparations:  
        prev_list = [FILEHANDLE_HDL] # solr always adds hdl before returning it!
        testcoupler = TESTHELPERS.get_coupler()
        TESTHELPERS.patch_solr_returns_previous_files(testcoupler, prev_list) # solr returns file list
        TESTHELPERS.patch_with_rabbit_mock(testcoupler)
        dsargs = TESTHELPERS.get_args_for_publication_assistant()
        assistant = DatasetPublicationAssistant(coupler=testcoupler, **dsargs)

        # Run code to be tested:
        assistant.add_file(**fileargs)
        assistant.dataset_publication_finished()

        # Check result (dataset):
        received_rabbit_task = TESTHELPERS.get_received_message_from_rabbitmock(testcoupler, 0)
        expected_rabbit_task = TESTHELPERS.get_rabbit_message_publication_dataset()
        expected_rabbit_task['files'] = [FILEHANDLE_HDL]
        same = utils.is_json_same(expected_rabbit_task, received_rabbit_task)
        self.assertTrue(same, error_message(expected_rabbit_task, received_rabbit_task))

        # Check result (file):
        received_rabbit_task = TESTHELPERS.get_received_message_from_rabbitmock(testcoupler, 1)
        expected_rabbit_task = TESTHELPERS.get_rabbit_message_publication_file()
        same = utils.is_json_same(expected_rabbit_task, received_rabbit_task)
        self.assertTrue(same, error_message(expected_rabbit_task, received_rabbit_task))

    def test_normal_publication_with_neg_consis_check_missing_file(self):

        # Test variables:
        args1 = TESTHELPERS.get_args_for_adding_file()
        prev1 = args1['file_handle']
        prev2 = PREFIX_NO_HDL +'/random_suffix_abc123'
        prev_list = [prev1, prev2]

        # Preparations:
        testcoupler = TESTHELPERS.get_coupler()
        TESTHELPERS.patch_solr_returns_previous_files(testcoupler, prev_list) # solr returns file list
        TESTHELPERS.patch_with_rabbit_mock(testcoupler)
        dsargs = TESTHELPERS.get_args_for_publication_assistant()
        assistant = DatasetPublicationAssistant(coupler=testcoupler, **dsargs)

        # Prepare:
        assistant.add_file(**args1)

        # Run code to be tested and check exception:
        with self.assertRaises(esgfpid.exceptions.InconsistentFilesetException):
            assistant.dataset_publication_finished()

    def test_normal_publication_with_neg_consis_check_too_many_files(self):

        # Test variables:
        args1 = TESTHELPERS.get_args_for_adding_file()
        args2 = TESTHELPERS.get_args_for_adding_file()
        handle1 = args1['file_handle']
        handle2 = PREFIX_NO_HDL +'/random_suffix_abc123'
        args2['file_handle'] = handle2
        prev_list = [handle1]

        # Preparations:
        testcoupler = TESTHELPERS.get_coupler()
        TESTHELPERS.patch_with_rabbit_mock(testcoupler)
        TESTHELPERS.patch_solr_returns_previous_files(testcoupler, prev_list) # solr returns file list
        dsargs = TESTHELPERS.get_args_for_publication_assistant()
        assistant = DatasetPublicationAssistant(coupler=testcoupler, **dsargs)

        # Prepare:
        assistant.add_file(**args1)
        assistant.add_file(**args2)

        # Run code to be tested and check exception:
        with self.assertRaises(esgfpid.exceptions.InconsistentFilesetException):
            assistant.dataset_publication_finished()

    #
    # Testing invalid operations
    # (publication actions occur at the wrong time / wrong state)
    #

    def test_add_file_too_late(self):

        # Prepare:
        testcoupler = TESTHELPERS.get_coupler(solr_switched_off=True) # solr switched off, no consistency check
        TESTHELPERS.patch_with_rabbit_mock(testcoupler)
        dsargs = TESTHELPERS.get_args_for_publication_assistant()
        assistant = DatasetPublicationAssistant(coupler=testcoupler, **dsargs)
        fileargs = TESTHELPERS.get_args_for_adding_file()
        assistant.add_file(**fileargs)
        assistant.dataset_publication_finished()

        # Run code to be tested and check exception:
        with self.assertRaises(esgfpid.exceptions.OperationUnsupportedException):
            assistant.add_file(**fileargs)

    def test_finish_too_late(self):

        # Prepare:
        testcoupler = TESTHELPERS.get_coupler(solr_switched_off=True) # solr switched off, no consistency check
        TESTHELPERS.patch_with_rabbit_mock(testcoupler)
        dsargs = TESTHELPERS.get_args_for_publication_assistant()
        assistant = DatasetPublicationAssistant(coupler=testcoupler, **dsargs)
        fileargs = TESTHELPERS.get_args_for_adding_file()
        assistant.add_file(**fileargs)
        assistant.dataset_publication_finished()

        # Run code to be tested:
        with self.assertRaises(esgfpid.exceptions.OperationUnsupportedException):
            assistant.dataset_publication_finished()

    def test_finish_too_early(self):

        # Test variables:
        testcoupler = TESTHELPERS.get_coupler(solr_switched_off=True) # solr switched off, no consistency check
        dsargs = TESTHELPERS.get_args_for_publication_assistant()
        assistant = DatasetPublicationAssistant(coupler=testcoupler, **dsargs)

        # Run code to be tested and check exception:
        with self.assertRaises(esgfpid.exceptions.OperationUnsupportedException):
            assistant.dataset_publication_finished()

    #
    # Testing replica publication
    #

    def test_normal_publication_replica_flag_bool_ok(self):

        # Preparations:
        testcoupler = TESTHELPERS.get_coupler(solr_switched_off=True) # solr switched off, no consistency check
        TESTHELPERS.patch_with_rabbit_mock(testcoupler)
        dsargs = TESTHELPERS.get_args_for_publication_assistant()
        dsargs['is_replica'] = True
        assistant = DatasetPublicationAssistant(coupler=testcoupler, **dsargs)
        fileargs = TESTHELPERS.get_args_for_adding_file()

        # Run code to be tested:
        assistant.add_file(**fileargs)
        assistant.dataset_publication_finished()

        # Check result (dataset):
        received_rabbit_task = TESTHELPERS.get_received_message_from_rabbitmock(testcoupler, 0)
        expected_rabbit_task = TESTHELPERS.get_rabbit_message_publication_dataset()
        expected_rabbit_task['is_replica'] = True
        expected_rabbit_task['ROUTING_KEY'] = PREFIX_FOR_ROUTINGKEY+'.HASH.fresh.publi-ds-repli'
        same = utils.is_json_same(expected_rabbit_task, received_rabbit_task)
        self.assertTrue(same, error_message(expected_rabbit_task, received_rabbit_task))

        # Check result (file):
        received_rabbit_task = TESTHELPERS.get_received_message_from_rabbitmock(testcoupler, 1)
        expected_rabbit_task = TESTHELPERS.get_rabbit_message_publication_file()
        expected_rabbit_task['is_replica'] = True
        expected_rabbit_task['ROUTING_KEY'] = PREFIX_FOR_ROUTINGKEY+'.HASH.fresh.publi-file-repli'
        same = utils.is_json_same(expected_rabbit_task, received_rabbit_task)
        self.assertTrue(same, error_message(expected_rabbit_task, received_rabbit_task))

    def test_normal_publication_replica_flag_string_ok(self):

        # Preparations:
        testcoupler = TESTHELPERS.get_coupler(solr_switched_off=True) # solr switched off, no consistency check
        TESTHELPERS.patch_with_rabbit_mock(testcoupler)
        dsargs = TESTHELPERS.get_args_for_publication_assistant()
        dsargs['is_replica'] = True
        assistant = DatasetPublicationAssistant(coupler=testcoupler, **dsargs)
        fileargs = TESTHELPERS.get_args_for_adding_file()

        # Run code to be tested:
        assistant.add_file(**fileargs)
        assistant.dataset_publication_finished()

        # Check result (dataset):
        received_rabbit_task = TESTHELPERS.get_received_message_from_rabbitmock(testcoupler, 0)
        expected_rabbit_task = TESTHELPERS.get_rabbit_message_publication_dataset()
        expected_rabbit_task['is_replica'] = True
        expected_rabbit_task['ROUTING_KEY'] = PREFIX_FOR_ROUTINGKEY+'.HASH.fresh.publi-ds-repli'
        same = utils.is_json_same(expected_rabbit_task, received_rabbit_task)
        self.assertTrue(same, error_message(expected_rabbit_task, received_rabbit_task))

        # Check result (file):
        received_rabbit_task = TESTHELPERS.get_received_message_from_rabbitmock(testcoupler, 1)
        expected_rabbit_task = TESTHELPERS.get_rabbit_message_publication_file()
        expected_rabbit_task['is_replica'] = True
        expected_rabbit_task['ROUTING_KEY'] = PREFIX_FOR_ROUTINGKEY+'.HASH.fresh.publi-file-repli'
        same = utils.is_json_same(expected_rabbit_task, received_rabbit_task)
        self.assertTrue(same, error_message(expected_rabbit_task, received_rabbit_task))

    #
    # Helper tests
    #

    def test_get_dataset_handle(self):

        # Preparations:
        testcoupler = TESTHELPERS.get_coupler(solr_switched_off=True) # solr switched off, no consistency check
        dsargs = TESTHELPERS.get_args_for_publication_assistant()
        dsargs['is_replica'] = True
        assistant = DatasetPublicationAssistant(coupler=testcoupler, **dsargs)

        # Run code to be tested:
        handle = assistant.get_dataset_handle()

        # Check result:
        exp_handle = DATASETHANDLE_HDL
        self.assertEqual(handle, exp_handle,
            'Wrong handle returned.')


'''
This module is responsible for all the communication with the RabbitMQ
messaging service. It provides one class, the "AsynchronousRabbitConnector".

It communicates with the server in an asychronous way, by opening
another thread which runs in parallel, waits for events from the library
(e.g. message publication requests) and from the RabbitMQ (e.g. message
receival confirmations).

After initialisation, it is necessary to start the thread by calling
"start_messaging_thread()". This opens the parallel thread and triggers 
the connection to the server.

Then, the messages can be sent using "send_message_to_queue()" or
"send_many_messages()".

After the messaging business is done, it is necessary to close the
thread by calling "finish_rabbit_thread()" or "force_finish_rabbit_thread()".

  .. note:: Please note that KeyboardInterrupts may not work while a
    second thread is running. They need to be explicitly caught in
    the code calling the esgfpid library.

'''

import sys
if sys.version[0] == '2':
    import Queue as queue
else:
    import queue as queue
import threading
import pika
import time
import json
import copy
import datetime
import logging
import esgfpid.utils
from esgfpid.utils import loginfo, logdebug, logtrace, logerror, logwarn, log_every_x_times
from .rabbitthread import RabbitThread
from .thread_statemachine import StateMachine
from .exceptions import OperationNotAllowed

LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())


class AsynchronousRabbitConnector(object):

    '''
    Constructor for the asychronous rabbit connection module.

    This does not open a connection or start a thread yet.

    :param node_manager: NodeManager object that contains 
        the info about all the available RabbitMQ instances,
        their credentials, their priorities.

    '''
    def __init__(self, node_manager):
        logdebug(LOGGER, 'Initializing rabbit connector...')

        '''
        To check whether the thread has been started yet.
        If not, the methods of this module will raise an error, to
        make sure that the library caller will start the thread.
        '''
        self.__not_started_yet = True

        # To be filled after join:
        self.__leftovers_unpublished = [] # will be filled after join
        self.__leftovers_unconfirmed = [] # will be filled after join
        self.__leftovers_nacked      = [] # will be filled after join

        # Shared objects
        self.__statemachine = StateMachine()
        self.__unpublished_messages_queue = queue.Queue()

        # Log flags
        self.__first_message_receival = True
        self.__logcounter_received = 1
        self.__LOGFREQUENCY = 10

        # Actually created the thread:
        #self.__thread = RabbitThread(self.__statemachine, self.__unpublished_messages_queue, self, node_manager)
        self.__thread = self.__create_thread(node_manager)

        logdebug(LOGGER, 'Initializing rabbit connector... done.')

    def __create_thread(self, node_manager): # easy to mock/patch in unit test!
        return RabbitThread(self.__statemachine, self.__unpublished_messages_queue, self, node_manager)


    '''
    This starts the parallel thread.
    Needs to be called manually by library user.

    We would like to encourage the library user to start the
    thread as early as possible, even before using the library
    to publish messages.
    This is only possible if the thread can be started (and
    stopped) separately from the rest of the library's functionality.

    Reasons for this method not to be called during init of
    the library may be:

    * Some usage may not need any connection, e.g. if the library is only used for creating handle strings.
    * Testing of the library is easier.
    * If starting needs to be called explicitly, it is less likely that the stopping is forgotten, so abandoned connections and unjoined threads are avoided.

    '''
    def start_rabbit_thread(self):
        self.__not_started_yet = False
        self.__statemachine.set_to_waiting_to_be_available()
        self.__thread.start()

    #################
    ### Finishing ###
    #################

    def is_finished(self):
        return (not self.__thread.is_alive())

    '''
    "Gentle" finish of the thread.
    If any messages are not published or confirmed yet, the
    thread waits and rechecks for some time.

    So this message may block while the rabbit thread waits
    for some pending messages, and it may block while the
    main thread tries to join the rabbit thread.
    '''
    def finish_rabbit_thread(self):
        logdebug(LOGGER, 'Finishing...')

        # Make sure no more messages are accepted from publisher
        # while publishes/confirms are still accepted:
        if self.__statemachine.is_AVAILABLE():
            self.__statemachine.set_to_wanting_to_stop()
        self.__statemachine.detail_asked_to_gently_close_by_publisher = True

        # Asking the thread to finish:
        self.__thread.add_event_gently_finish() # (this blocks!)
        logdebug(LOGGER, 'Finishing... done')
        self.__join_and_rescue()

    '''
    Forces the immediate close-down of the thread, no matter
    if messages are still pending (i.e. not published or not
    confirmed).

    Note: If this method blocks, it is because the joining
    of the thread blocks.
    '''
    def force_finish_rabbit_thread(self):
        logdebug(LOGGER, 'Force finishing...')

        # Make sure no more messages are accepted from publisher
        # while confirms are still accepted:
        self.__statemachine.detail_asked_to_force_close_by_publisher = True
        self.__statemachine.set_to_wanting_to_stop()

        # Asking the thread to finish:
        logwarn(LOGGER, 'Forced close down of message sending module. Will not wait for pending messages, if any.')
        self.__thread.add_event_force_finish()
        logdebug(LOGGER, 'Force finishing... done')
        self.__join_and_rescue()

    '''
    Tries several times to join the thread.
    Note: May block for up to 6 seconds!
    '''
    def __join_and_rescue(self):
        timeout_seconds = 2
        success = self.__join(timeout_seconds)
        if success:
            self.__rescue_leftovers()
        else:
            loginfo(LOGGER, 'Joining the thread failed once... Retrying.')
            for i in range(20):
                time.sleep(0.1) # blocking
            self.__thread.add_event_force_finish()
            success = self.__join(timeout_seconds)
            if success:
                self.__rescue_leftovers()
            else:
                logerror(LOGGER, 'Joining failed again. No idea why.')

    def __join(self, timeout_seconds):        
        logdebug(LOGGER, 'Joining...')
        self.__thread.join(timeout_seconds)
        if self.__thread.is_alive():
            logdebug(LOGGER, 'Joining failed.')
            return False
        else:
            logdebug(LOGGER, 'Joining... done')
            return True  

    ##############################################
    ### Unpublished/unconfimed/nacked messages ###
    ##############################################

    '''
    Stores the pending messages in variables after
    closing the rabbit thread.
    So far, nothing is done with them, as republishing
    them does not work after closing, and writing to
    console or to file makes no sense.
    '''
    def __rescue_leftovers(self):
        logdebug(LOGGER, 'Storing unpublished/unconfirmed messages...')
        self.__rescue_unpublished_messages()
        self.__rescue_nacked_messages()
        self.__rescue_unconfirmed_messages()
        logdebug(LOGGER, 'Storing unpublished/unconfirmed messages... done.')      

    def __rescue_unpublished_messages(self):
        if self.__thread.is_alive():
            logwarn(LOGGER, 'Cannot retrieve unpublished messages while thread still alive.')
        else:
            self.__leftovers_unpublished = self.__get_unpublished_messages_as_list()
            num = len(self.__leftovers_unpublished)
            if num > 0:
                logdebug(LOGGER, 'Rescued %i unpublished messages.', num)
            else:
                logdebug(LOGGER, 'No unpublished messages to rescue.')

    def __rescue_unconfirmed_messages(self):
        if self.__thread.is_alive():
            logwarn(LOGGER, 'Cannot retrieve unconfirmed messages while thread still alive.')
        else:
            self.__leftovers_unconfirmed = self.__get_unconfirmed_messages_as_list()
            num = len(self.__leftovers_unconfirmed)
            if num > 0:
                logdebug(LOGGER, 'Rescued %i unconfirmed messages.', num)
            else:
                logdebug(LOGGER, 'No unconfirmed messages to rescue.')

    def __rescue_nacked_messages(self):
        if self.__thread.is_alive():
            logwarn(LOGGER, 'Cannot retrieve rejected (NACKed) messages while thread still alive.')
        else:
            self.__leftovers_nacked = self.__get_nacked_messages_as_list()
            num = len(self.__leftovers_nacked)
            if num > 0:
                logdebug(LOGGER, 'Rescued %i rejected (NACKed) messages.', num)
            else:
                logdebug(LOGGER, 'No rejected (NACKed) messages to rescue.')

    '''
    This code is not currently used, as the library does not provide
    any possibility to reconnect manually.
    The republication of unpublished/unconfirmed messages after a
    connection interruption is handled by the RabbitThread itself.

    def __publish_leftovers(self):
        self.__publish_leftovers_unpublished()
        self.__publish_leftovers_unconfirmed()
        self.__publish_leftovers_nacked()

    def __publish_leftovers_unpublished(self):
        if len(self.__leftovers_unpublished) > 0:
            self.__thread.send_many_messages(self.__leftovers_unpublished)
            self.__leftovers_unpublished = []

    def __publish_leftovers_unconfirmed(self):
        if len(self.__leftovers_unconfirmed) > 0:
            self.__thread.send_many_messages(self.__leftovers_unconfirmed)
            self.__leftovers_unconfirmed = []

    def __publish_leftovers_nacked(self):
        if len(self.__leftovers_nacked) > 0:
            self.__thread.send_many_messages(self.__leftovers_nacked)
            self.__leftovers_nacked = []
    '''

    #def any_leftovers(self):
    #    if (len(self.__leftovers_unpublished)+
    #        len(self.__leftovers_unconfirmed)+
    #        len(self.__leftovers_nacked))>0:
    #        return True
    #    return False

    #def get_leftovers(self):
    #    return self.__leftovers_unpublished + self.__leftovers_unconfirmed + self.__leftovers_nacked


    ####################
    # Sending messages #
    ####################

    '''
    Send a JSON message to RabbitMQ.

    :param message: JSON message to be published.
    :raises: OperationNotAllowed: If the rabbit thread was not started yet
    or stopped again.
    '''
    def send_message_to_queue(self, message):
        if self.__not_started_yet:
            msg = ('Cannot publish message. The message sending module was not initalized yet. '+
                   '(Please call the PID connector\'s "start_messaging_thread()" before trying '+
                   'to send messages, and do not forget to "finish_messaging_thread()" afterwards.')
            raise OperationNotAllowed(msg)
            # Note: This exception is only thrown if the code that calls the library does
            # it wrong. So we throw this exception to remind the developer to start the rabbit
            # before using it.
        self.__send_a_message(message)

    '''
    Send many JSON messages to RabbitMQ.

    :param list_of_messages: List of JSON message to be published.
    :raises: OperationNotAllowed: If the rabbit thread was not started yet
    or stopped again.
    '''
    def send_many_messages_to_queue(self, list_of_messages):
        if self.__not_started_yet:
            msg = ('Cannot publish message. The message sending module was not initalized yet. '+
                   '(Please call the PID connector\'s "start_messaging_thread()" before trying '+
                   'to send messages, and do not forget to "finish_messaging_thread()" afterwards.')
            raise OperationNotAllowed(msg)
        self.__send_many_messages(list_of_messages)

    def __send_a_message(self, message):
        if self.__statemachine.is_WAITING_TO_BE_AVAILABLE():
            self.__log_receival_one_message(message)
            self.__put_one_message_into_queue_of_unsent_messages(message)

        elif self.__statemachine.is_AVAILABLE():
            self.__log_receival_one_message(message)
            self.__put_one_message_into_queue_of_unsent_messages(message)
            self.__trigger_one_publish_action()

        elif self.__statemachine.is_AVAILABLE_BUT_WANTS_TO_STOP() or self.__statemachine.is_PERMANENTLY_UNAVAILABLE() or self.__statemachine.is_FORCE_FINISHED():
            errormsg = 'Accepting no more messages'
            logdebug(LOGGER, errormsg+' (dropping %s).', message)
            logwarn(LOGGER, 'RabbitMQ module was closed and does not accept any more messages. Dropping message. Reason: %s', self.__statemachine.get_reason_shutdown())
            # Note: This may happen if the connection failed. We may not stop
            # the publisher in this case, so we do not raise an exception.
            # We only raise an exception if the closing was asked by the publisher!
            if self.__statemachine.get_detail_closed_by_publisher():
                raise OperationNotAllowed(errormsg)

        elif self.__statemachine.is_NOT_STARTED_YET():
            errormsg = 'Cannot send a message, the messaging thread was not started yet!'
            logwarn(LOGGER, errormsg+' (dropping %s).', message)
            raise OperationNotAllowed(errormsg)
            # This is almost the same as the one raised if self.__not_started_yet is True.


    def __send_many_messages(self, messages):
        if self.__statemachine.is_WAITING_TO_BE_AVAILABLE():
            self.__log_receival_many_messages(messages)
            self.__put_all_messages_into_queue_of_unsent_messages(messages)

        elif self.__statemachine.is_AVAILABLE():
            self.__log_receival_many_messages(messages)
            self.__put_all_messages_into_queue_of_unsent_messages(messages)
            self.__trigger_n_publish_actions(len(messages))

        elif self.__statemachine.is_AVAILABLE_BUT_WANTS_TO_STOP() or self.__statemachine.is_PERMANENTLY_UNAVAILABLE() or self.__statemachine.is_FORCE_FINISHED():
            errormsg = 'Accepting no more messages'
            logwarn(LOGGER, errormsg+' (dropping %i messages).', len(messages))
            if self.__statemachine.get_detail_closed_by_publisher():
                raise OperationNotAllowed(errormsg)

        elif self.__statemachine.is_NOT_STARTED_YET():
            errormsg = 'Cannot send any messages, the messaging thread was not started yet!'
            logwarn(LOGGER, errormsg+' (dropping %i messages).', len(messages))
            raise OperationNotAllowed(errormsg)

    def __put_one_message_into_queue_of_unsent_messages(self, message):
        logtrace(LOGGER, 'Putting a message into stack that waits to be published...')
        self.__unpublished_messages_queue.put(message, block=False)

    def __log_receival_one_message(self, message):
        if self.__first_message_receival:
            logdebug(LOGGER, 'Handing over first message to rabbit thread...')
            self.__first_message_receival = False
        logtrace(LOGGER, 'Handing over one message over to the rabbit thread (%s)', message)
        log_every_x_times(LOGGER, self.__logcounter_received, self.__LOGFREQUENCY, 'Handing over one message over to the rabbit thread (no. %i).', self.__logcounter_received)
        self.__logcounter_received += 1

    def __log_receival_many_messages(self, messages):
        if self.__first_message_receival:
            logdebug(LOGGER, 'Handing over first message to rabbit thread...')
            self.__first_message_receival = False
        logdebug(LOGGER, 'Batch sending: Handing %i messages over to the sender.', len(messages))
        self.__logcounter_received += len(messages)

    def __put_all_messages_into_queue_of_unsent_messages(self, messages):
        counter = 1
        for message in messages:
            logtrace(LOGGER, 'Adding message %i/%i to stack to be sent.', counter, len(messages))
            counter += 1
            self.__put_one_message_into_queue_of_unsent_messages(message)

    def __trigger_one_publish_action(self):
        self.__thread.add_event_publish_message()

    def __trigger_n_publish_actions(self, num_messages_to_publish):
        logdebug(LOGGER, 'Asking rabbit thread to publish %i messages...', num_messages_to_publish)
        to_be_sure = 10
        for i in range(num_messages_to_publish+to_be_sure):
            self.__thread.add_event_publish_message()


    ###############
    ### Getters ###
    ###############

    '''
    Returns a copy of the messages that were not confirmed during
    the rabbit thread's lifetime.
    Can only be called after the thread's join, as dict objects
    are not thread-safe.

    :raises: OperationNotAllowed: If the thread is still alive.
    :return: A list containing all unconfirmed messages.
    '''
    def __get_unconfirmed_messages_as_list(self):
        return self.__thread.get_unconfirmed_messages_as_list_copy()

    '''
    Returns a copy of the messages that were nacked during
    the rabbit thread's lifetime.
    Can only be called after the thread's join, as dict objects
    are not thread-safe.

    :raises: OperationNotAllowed: If the thread is still alive.
    :return: A list containing all nacked messages.
    '''
    def __get_nacked_messages_as_list(self):
        return self.__thread.get_nacked_messages_as_list()

    '''
    Returns a new list containing all unpublished messages.

    Note: As this method gets the content of a thread-safe queue,
    there is no harm in calling it during the rabbit thread's lifetime.
    However, we restrict its use to after its joining, because we
    only want to use it to "rescue" messages that have not been
    published once publishing definitely has ended.

    The method iterates over the Queue and retrieves objects until
    a queue.Empty event occurs. As this event is not guaranteed
    to be true, the retrieval is then tried another time with
    a little timeout to make sure all elements are retrieved.

    This complicated way of making a copy is not necessary
    if the method is only called after the thread's join...

    :raises: OperationNotAllowed: If the thread is still alive.
    :return: A list containing all unpublished messages.
    '''
    def __get_unpublished_messages_as_list(self):
        newlist = []
        counter = 0
        to_be_safe = 10
        max_iterations = self.__unpublished_messages_queue.qsize() + to_be_safe
        while True:
            counter+=1
            if counter == max_iterations:
                break
            try:
                self.__get_msg_from_queue_and_store_first_try(
                    newlist,
                    self.__unpublished_messages_queue)
            except queue.Empty:
                try:
                    self.__get_a_msg_from_queue_and_store_second_try(
                        newlist,
                        self.__unpublished_messages_queue)
                except queue.Empty:
                    break
        return newlist

    '''Put a message from the Queue to the list, without waiting.'''
    def __get_msg_from_queue_and_store_first_try(self, alist, msg_queue):
        msg_incl_routing_key = msg_queue.get(block=False)
        alist.append(msg_incl_routing_key)

    '''Put a message from the Queue to the list, with waiting.'''
    def __get_a_msg_from_queue_and_store_second_try(self, alist, msg_queue):
        wait_seconds = 0.5
        msg_incl_routing_key = msg_queue.get(block=True, timeout=wait_seconds)
        alist.append(msg_incl_routing_key)


import unittest
import mock
import logging
import json
import sys

LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())

# Test resources:
from resources.TESTVALUES import *
import resources.TESTVALUES as TESTHELPERS

'''
Tests the rabbit module's API (rabbit.py),
using a simple MagicMock for both synchronous and
asynchronous RabbitMQ server connector modules.
'''
class RabbitTestCase(unittest.TestCase):

    def setUp(self):
        LOGGER.info('######## Next test (%s) ##########', __name__)

    def tearDown(self):
        LOGGER.info('#############################')

    def test_constructor_syn(self,):

        # Test variables:
        args = TESTHELPERS.get_rabbit_args(is_synchronous_mode=True)

        # Run code to be tested:
        testrabbit = esgfpid.rabbit.RabbitMessageSender(**args)

        # Check result
        self.assertIsInstance(testrabbit, esgfpid.rabbit.RabbitMessageSender)
        serverconn = testrabbit._RabbitMessageSender__server_connector
        self.assertIsInstance(serverconn, esgfpid.rabbit.synchronous.SynchronousRabbitConnector,
            'Wrong type: %s' % type(serverconn))
        self.assertFalse(testrabbit._RabbitMessageSender__ASYNCHRONOUS)

    def test_constructor_asyn(self,):

        # Test variables:
        args = TESTHELPERS.get_rabbit_args(is_synchronous_mode=False)
        args['is_synchronous_mode'] = False

        # Run code to be tested:
        testrabbit = esgfpid.rabbit.RabbitMessageSender(**args)

        # Check result
        self.assertIsInstance(testrabbit, esgfpid.rabbit.RabbitMessageSender)
        serverconn = testrabbit._RabbitMessageSender__server_connector
        self.assertIsInstance(serverconn, esgfpid.rabbit.asynchronous.AsynchronousRabbitConnector,
            'Wrong type: %s' % type(serverconn))
        self.assertTrue(testrabbit._RabbitMessageSender__ASYNCHRONOUS)


    def test_open_rabbit_connection(self):

        # Make test rabbit:
        rabbit_syn  = TESTHELPERS.get_rabbit_message_sender(is_synchronous_mode=True)
        rabbit_asyn = TESTHELPERS.get_rabbit_message_sender(is_synchronous_mode=False)
        TESTHELPERS.patch_rabbit_with_magic_mock(rabbit_syn)
        TESTHELPERS.patch_rabbit_with_magic_mock(rabbit_asyn)

        # Run code to be tested:
        rabbit_syn.open_rabbit_connection()
        rabbit_asyn.open_rabbit_connection()

        # Check result:
        mock_connector_syn = rabbit_syn._RabbitMessageSender__server_connector
        mock_connector_syn.open_rabbit_connection.assert_called_with()
        mock_connector_asyn = rabbit_asyn._RabbitMessageSender__server_connector
        mock_connector_asyn.open_rabbit_connection.assert_not_called()

    def test_close_rabbit_connection(self):

        # Make test rabbit:
        rabbit_syn  = TESTHELPERS.get_rabbit_message_sender(is_synchronous_mode=True)
        rabbit_asyn = TESTHELPERS.get_rabbit_message_sender(is_synchronous_mode=False)
        TESTHELPERS.patch_rabbit_with_magic_mock(rabbit_syn)
        TESTHELPERS.patch_rabbit_with_magic_mock(rabbit_asyn)
        
        # Run code to be tested:
        rabbit_syn.close_rabbit_connection()
        rabbit_asyn.close_rabbit_connection()

        # Check result:
        mock_connector_syn = rabbit_syn._RabbitMessageSender__server_connector
        mock_connector_syn.close_rabbit_connection.assert_called_with()
        mock_connector_asyn = rabbit_asyn._RabbitMessageSender__server_connector
        mock_connector_asyn.close_rabbit_connection.assert_not_called()

    def test_send_message_to_queue(self):

        # Test variables
        msg = 'FOOBAR'

        # Make test rabbit and replace the connector with a mock:
        rabbit_syn  = TESTHELPERS.get_rabbit_message_sender(is_synchronous_mode=True)
        rabbit_asyn = TESTHELPERS.get_rabbit_message_sender(is_synchronous_mode=False)
        TESTHELPERS.patch_rabbit_with_magic_mock(rabbit_syn)
        TESTHELPERS.patch_rabbit_with_magic_mock(rabbit_asyn)

        # Run code to be tested
        rabbit_syn.send_message_to_queue(msg)
        rabbit_asyn.send_message_to_queue(msg)

        # Check result
        mock_connector = rabbit_syn._RabbitMessageSender__server_connector
        mock_connector.send_message_to_queue.assert_called_with(msg)
        mock_connector = rabbit_asyn._RabbitMessageSender__server_connector
        mock_connector.send_message_to_queue.assert_called_with(msg)


    def test_is_finished(self):

        # Make test rabbit and replace the connector with a mock:
        rabbit_syn  = TESTHELPERS.get_rabbit_message_sender(is_synchronous_mode=True)
        rabbit_asyn = TESTHELPERS.get_rabbit_message_sender(is_synchronous_mode=False)
        TESTHELPERS.patch_rabbit_with_magic_mock(rabbit_syn)
        TESTHELPERS.patch_rabbit_with_magic_mock(rabbit_asyn)

        # Run code to be tested
        res1 = rabbit_syn.is_finished()
        res2 = rabbit_asyn.is_finished()

        # Check result
        self.assertIsNone(res1)
        self.assertTrue(res2, 'We asked the mock to return True, not %s.' % res2)
        mock_connector_asyn = rabbit_asyn._RabbitMessageSender__server_connector
        mock_connector_asyn.is_finished.assert_called()

    def test_finish_ok(self):

        # Make test rabbit and replace the connector with a mock:
        rabbit_syn  = TESTHELPERS.get_rabbit_message_sender(is_synchronous_mode=True)
        rabbit_asyn = TESTHELPERS.get_rabbit_message_sender(is_synchronous_mode=False)
        TESTHELPERS.patch_rabbit_with_magic_mock(rabbit_syn)
        TESTHELPERS.patch_rabbit_with_magic_mock(rabbit_asyn)

        # Run code to be tested
        res1 = rabbit_syn.finish()
        res2 = rabbit_asyn.finish()


        # Check result
        mock_connector = rabbit_syn._RabbitMessageSender__server_connector
        mock_connector.finish_rabbit_thread.assert_not_called()
        mock_connector = rabbit_asyn._RabbitMessageSender__server_connector
        mock_connector.finish_rabbit_thread.assert_called_with()


    def test_force_finish_ok(self):

        # Make test rabbit and replace the connector with a mock:
        rabbit_syn  = TESTHELPERS.get_rabbit_message_sender(is_synchronous_mode=True)
        rabbit_asyn = TESTHELPERS.get_rabbit_message_sender(is_synchronous_mode=False)
        TESTHELPERS.patch_rabbit_with_magic_mock(rabbit_syn)
        TESTHELPERS.patch_rabbit_with_magic_mock(rabbit_asyn)

        # Run code to be tested
        res1 = rabbit_syn.force_finish()
        res2 = rabbit_asyn.force_finish()

        # Check result
        mock_connector = rabbit_syn._RabbitMessageSender__server_connector
        mock_connector.force_finish_rabbit_thread.assert_not_called()
        mock_connector = rabbit_asyn._RabbitMessageSender__server_connector
        mock_connector.force_finish_rabbit_thread.assert_called_with()


    def test_start_ok(self):

        # Make test rabbit and replace the connector with a mock:
        rabbit_syn  = TESTHELPERS.get_rabbit_message_sender(is_synchronous_mode=True)
        rabbit_asyn = TESTHELPERS.get_rabbit_message_sender(is_synchronous_mode=False)
        TESTHELPERS.patch_rabbit_with_magic_mock(rabbit_syn)
        TESTHELPERS.patch_rabbit_with_magic_mock(rabbit_asyn)

        # Run code to be tested
        res1 = rabbit_syn.start()
        res2 = rabbit_asyn.start()

        # Check result
        mock_connector = rabbit_syn._RabbitMessageSender__server_connector
        mock_connector.start_rabbit_thread.assert_not_called()
        mock_connector = rabbit_asyn._RabbitMessageSender__server_connector
        mock_connector.start_rabbit_thread.assert_called_with()


    def test_leftovers_ok(self):
        
        # Make test rabbit and replace the connector with a mock:
        rabbit_asyn = TESTHELPERS.get_rabbit_message_sender(is_synchronous_mode=False)
        TESTHELPERS.patch_rabbit_with_magic_mock(rabbit_asyn)

        # Replace the connector with a mock:
        mock_connector = rabbit_asyn._RabbitMessageSender__server_connector
        mock_connector.any_leftovers.return_value = True
        mock_connector.get_leftovers.return_value = ['a', 'b', 'c']
        
        # Run code to be tested
        result_bool = rabbit_asyn.any_leftovers()
        result_list = rabbit_asyn.get_leftovers()

        # Check result:
        self.assertTrue(result_bool,
            'We asked the mock to return True, not %s.' % result_bool)
        self.assertEqual(result_list, ['a', 'b', 'c'],
            'We asked the mock to return a list, not %s.' % result_list)
        mock_connector.any_leftovers.assert_called_with()
        mock_connector.get_leftovers.assert_called_with()


import ssl
import pika
import copy
import logging
import random
import esgfpid.defaults
import esgfpid.exceptions
from esgfpid.utils import loginfo, logdebug, logtrace, logerror, logwarn, log_every_x_times, make_logsafe
from .naturalsorting import natural_keys
import esgfpid.utils

LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())


DEFAULT_PRIO = 'zzz_default'
LAST_PRIO = 'zzzz_last'

'''
This class is responsible for keeping track of RabbitMQ
instances and providing tha access info to the library.

At startup, it is fed all the info about the various
instances, using "add_trusted_node()" and "add_open_node()".
During the library's functioning, it provides the info to
the library's RabbitMQ connection modules.

On every call of 

It is fed by all the RabbitMQ instances, their
access info and priority, in the beginning. Then it is
responsible for providing this info to the library.

It can deal with trusted and open nodes and with integer
priorities. It returns the instance access info dictionaries
in a well-defined order, 
'''
class NodeManager(object):

    '''
    Constructor that takes no params. It creates an empty
    container for RabbitMQ node information. The node 
    information then has to be added using "add_trusted_node()"
    and "add_open_node()".
    '''
    def __init__(self):

        # Props for basic_publish (needed by thread_feeder)
        self.__properties = pika.BasicProperties(
            delivery_mode=esgfpid.defaults.RABBIT_DELIVERY_MODE,
            content_type='application/json',
        )

        # Nodes
        self.__trusted_nodes = {}
        self.__open_nodes = {}
        self.__trusted_nodes_archive = {}
        self.__open_nodes_archive = {}
        # Each of these dictionaries has the priorities as keys (integers
        # stored as strings, or 'zzz_default', if no prio was given, or
        # 'zzzz_last' if the host had failed in the pre-flight check).
        # For each priority, there is a list of node-info-dictionaries:
        # self.__trusted_nodes = {
        #     "1":         [node_info1, node_info2],
        #     "2":         [node_info3],
        #     "zzz_default": [node_info4]
        # }

        # Current node
        self.__current_node = None
        self.__exchange_name = None

        # Important info
        self.__has_trusted = False

    '''
    Add information about a trusted RabbitMQ node to
    the container, for later use.

    :param username: The username to connect to RabbitMQ.
    :param password: The password to connect to RabbitMQ.
    :param host: The host name of the RabbitMQ instance.
    :param exchange_name: The  exchange to which to send the
                          messages.
    :param priority: Optional. Integer priority for the use
                     of this instance.
    '''
    def add_trusted_node(self, **kwargs):
        kwargs['is_open'] = False
        node_info = self.__add_node(self.__trusted_nodes, self.__trusted_nodes_archive, **kwargs)
        self.__has_trusted = True
        logdebug(LOGGER, 'Trusted rabbit: %s', self.__get_node_log_string(node_info))

    '''
    Add information about an open RabbitMQ node to
    the container, for later use.

    The parameters that are needed are the same as
    for trusted nodes.

    Note that a password is needed!

    :param username: The username to connect to RabbitMQ.
    :param password: The password to connect to RabbitMQ.
    :param host: The host name of the RabbitMQ instance.
    :param exchange_name: The  exchange to which to send the
                          messages.
    :param priority: Optional. Integer priority for the use
                     of this instance.
    '''
    def add_open_node(self, **kwargs):
        raise esgfpid.exceptions.ArgumentError('Open nodes no longer supported! (Messaging service "'+kwargs['host']+'")')
        #kwargs['is_open'] = True
        #added = node_info = self.__add_node(self.__open_nodes, self.__open_nodes_archive, **kwargs)
        #logdebug(LOGGER, 'Open rabbit: %s', self.__get_node_log_string(node_info))

    def __add_node(self, store_where, store_archive, **kwargs):
        if self.__has_necessary_info(kwargs):
            node_info = copy.deepcopy(kwargs)
            node_copy = copy.deepcopy(kwargs)
            self.__complete_info_dict(node_info, kwargs['is_open'])
            self.__complete_info_dict(node_copy, kwargs['is_open'])
            self.__store_node_info_by_priority(node_info, store_where)
            self.__store_node_info_by_priority(node_copy, store_archive)
            #store_where[node_info['priority']].append(node_info)
            #store_archive[node_info['priority']].append(copy.deepcopy(node_info))
            return node_info
        else:
            kwargs_logsafe = make_logsafe(kwargs)
            raise esgfpid.exceptions.ArgumentError('Cannot add this RabbitMQ node. Missing info. Required: username, password, host and exchange_name. Provided: '+str(kwargs_logsafe))

    def __compare_nodes(self, cand1, cand2):
        # These cannot be compared by "==".
        # They are created from the other info, so neglecting
        # them in this comparison is ok!
        ignore_keys = ['credentials', 'params']
        copy1 = dict((k, v) for k,v in cand1.items() if k not in ignore_keys)
        copy2 = dict((k, v) for k,v in cand2.items() if k not in ignore_keys)
        return copy1 == copy2

    def __is_this_node_in_last_prio_already(self, where_to_look):
        try:
            list_candidates = where_to_look[LAST_PRIO]
        except KeyError as e:
            errmsg = 'No node of last prio (%s) exists.' % LAST_PRIO
            logwarn(LOGGER, errmsg)
            return False

        for i in range(len(list_candidates)):
            candidate = list_candidates[i]
            if self.__compare_nodes(candidate,self.__current_node):
                logtrace(LOGGER, 'Found current node in archive (in list of last-prio nodes).')
                return True

        return False

    def __move_to_last_prio(self, current_prio, all_nodes):

        list_candidates = all_nodes[current_prio]
        logsafe = make_logsafe(list_candidates)
        loginfo(LOGGER, 'Nodes of prio "%s": %s', current_prio, logsafe)

        for i in range(len(list_candidates)):
            candidate = list_candidates[i]
            if self.__compare_nodes(candidate,self.__current_node):
                logtrace(LOGGER, 'Found current node in archive.')

                # Add to lowest prio:
                try:
                    all_nodes[LAST_PRIO].append(candidate)
                    logdebug(LOGGER, 'Added this host to list of lowest prio hosts...')

                except KeyError:
                    all_nodes[LAST_PRIO] = [candidate]
                    logdebug(LOGGER, 'Added this host to (newly-created) list of lowest prio hosts...')

                # Remove from current prio:
                list_candidates.pop(i)
                loginfo(LOGGER, 'Removed this host from list of hosts with prio %s!', current_prio)
                if len(list_candidates)==0:
                    all_nodes.pop(current_prio)
                    loginfo(LOGGER, 'Removed the current priority %s!', current_prio)
                return True

        return False



    def set_priority_low_for_current(self):
        # We do not change the priority stored ass attribute in the
        # dicts, BUT we change the priority under which it is stored in
        # the list of nodes to be used.

        # Deal with open or trusted node:
        if self.__current_node['is_open']:
            where_to_look = self.__open_nodes_archive
        else:
            where_to_look = self.__trusted_nodes_archive

        # Go over all nodes of the current prio to find the
        # current one, then move it to a different prio:
        moved = False
        try:
            current_prio = self.__current_node['priority']
            moved = self.__move_to_last_prio(current_prio, where_to_look)
            if moved: return # changed successfully!

        except KeyError as e:
            logsafe = make_logsafe(where_to_look)
            errmsg = 'No node of prio %s found. Nodes: %s.' % (current_prio, logsafe)
            logwarn(LOGGER, errmsg)

            # The node had already been added to the last-prio nodes ?!
            last_already = self.__is_this_node_in_last_prio_already(where_to_look)
            if last_already:
                logdebug(LOGGER, 'Node already had lowest priority.')
                return # nothing to change!

        # This is extremely unlikely - in fact I don't see how it could occur:
        if (not moved) and (not last_already):
            errmsg = 'Could not find this node\'s priority (%s), nor the last-priority (%s). Somehow this node\'s priority was changed weirdly.' % (current_prio, LAST_PRIO)
            logwarn(LOGGER, errmsg)
            logsafe = make_logsafe(where_to_look)
            logwarn(LOGGER, 'All nodes: %s' % logsafe)

            # No matter where the node is stored, move it to "last" prio:
            for prio, nodes in where_to_look.items():

                logtrace(LOGGER, 'Looking in prio "%s"...' % prio)
                moved = self.__move_to_last_prio(prio, where_to_look)
                if moved: return # changed successfully!

            errmsg = 'Node definitely not found, cannot change prio.'
            logwarn(LOGGER, errmsg)
            raise ValueError(errmsg)


    def __store_node_info_by_priority(self, node_info, store_where):
        try:
            store_where[node_info['priority']].append(node_info)
        except KeyError:
            store_where[node_info['priority']] = [node_info]

    def __get_node_log_string(self, node_info):
        return ('%s, %s, %s (exchange "%s")' % (node_info['host'], node_info['username'], node_info['password'], node_info['exchange_name']))

    def __has_necessary_info(self, node_info_dict):
        if ('username' in node_info_dict and
           'password' in node_info_dict and
           'host' in node_info_dict and 
           'exchange_name' in node_info_dict and
            node_info_dict['username'] is not None and
            node_info_dict['password'] is not None and
            node_info_dict['host'] is not None and
            node_info_dict['exchange_name'] is not None):
            return True
        else:
            return False

    def __complete_info_dict(self, node_info_dict, is_open):

        # Make pika credentials
        creds = pika.PlainCredentials(
            node_info_dict['username'],
            node_info_dict['password']
        )
        node_info_dict['credentials'] = creds
        if 'priority' in node_info_dict and node_info_dict['priority'] is not None:
            node_info_dict['priority'] = str(node_info_dict['priority'])
        else:
            node_info_dict['priority'] = DEFAULT_PRIO

        # Mandatories:
        host = node_info_dict['host']
        credentials = node_info_dict['credentials']

        # Optional ones
        # If not specified, fill in defaults.
        vhost = ""
        if 'vhost' in node_info_dict and node_info_dict['vhost'] is not None:
            vhost = node_info_dict['vhost']
        port = 15672
        if 'port' in node_info_dict and node_info_dict['port'] is not None:
            port = node_info_dict['port']
        ssl_enabled = False
        if 'ssl_enabled' in node_info_dict and node_info_dict['ssl_enabled'] is not None:
            ssl_enabled = node_info_dict['ssl_enabled']

        # Create SSLOptions if SSL is enabled
        ssl_options = None
        if ssl_enabled:
            ssl_options = pika.SSLOptions(ssl.create_default_context())
            # I did not test pika and SSL in python 3.7 yet! Issues were reported:
            # https://github.com/pika/pika/issues/1107
            # https://pika.readthedocs.io/en/stable/examples/tls_mutual_authentication.html
            # TODO Test and fix SSL in pika with python 3.7

        # Get some defaults:
        socket_timeout = esgfpid.defaults.RABBIT_PIKA_SOCKET_TIMEOUT
        connection_attempts = esgfpid.defaults.RABBIT_PIKA_CONNECTION_ATTEMPTS
        retry_delay = esgfpid.defaults.RABBIT_PIKA_CONNECTION_RETRY_DELAY_SECONDS
        
        # Make pika connection params
        # https://pika.readthedocs.org/en/0.9.6/connecting.html
        params = pika.ConnectionParameters(
            host=host,
            ssl_options=ssl_options,
            port=port,
            virtual_host=vhost,
            credentials=credentials,
            socket_timeout=socket_timeout,
            connection_attempts=connection_attempts,
            retry_delay=retry_delay
        )

        node_info_dict['params'] = params

        # Add some stuff
        node_info_dict['is_open'] = is_open
        '''
        https://pika.readthedocs.org/en/0.9.6/connecting.html
        class pika.connection.ConnectionParameters(
            host=None, port=None, virtual_host=None, credentials=None, channel_max=None,
            frame_max=None, heartbeat_interval=None, ssl=None, ssl_options=None,
            connection_attempts=None, retry_delay=None, socket_timeout=None, locale=None,
            backpressure_detection=None)
        # Note: heartbeat_interval renamed to heartbeat in pika 1.0.0
        '''
        return node_info_dict

    '''
    Return the connection parameters for the current
    RabbitMQ host.

    :return: Connection parameters (of type pika.ConnectionParameters)
    '''
    def get_connection_parameters(self):
        if self.__current_node is None:
            self.set_next_host()
        if self.__current_node['is_open']:
            raise ArgumentError('Open nodes no longer supported! (Messaging service "'+credentials['url']+'")')
        return self.__current_node['params']

    '''
    Simple getter to find out if any URLs are
    left.

    TODO: Needed for what, as we start over
    once all have been used? There is no end!

    :return: Boolean.
    '''

    def has_more_urls(self):
        if self.get_num_left_urls() > 0:
            return True
        return False

    '''
    Compute and return the number of trusted
    RabbitMQ instances.

    :return: Number of trusted instances (integer).
    '''
    def get_num_left_trusted(self):
        n_trusted = 0
        for list_of_nodes in list(self.__trusted_nodes.values()):
            n_trusted = n_trusted + len(list_of_nodes)
        return n_trusted

    '''
    Compute and return the number of open
    RabbitMQ instances.

    :return: Number of open instances (integer).
    '''
    def get_num_left_open(self):
        n_open = 0
        for list_of_nodes in list(self.__open_nodes.values()):
            n_open = n_open + len(list_of_nodes)
        return n_open

    '''
    Compute and return the total number of RabbitMQ
    instances.

    :return: Number of trusted instances (integer).
    '''
    def get_num_left_urls(self):
        return self.get_num_left_open() + self.get_num_left_trusted()


    '''
    Select the next RabbitMQ to be used, using the
    predefined priorities. It is not returned.
    '''
    def set_next_host(self):

        if len(self.__trusted_nodes) > 0:
            self.__current_node = self.__get_highest_priority_node(self.__trusted_nodes)
            logdebug(LOGGER, 'Selected a trusted node: %s', self.__current_node['host'])

        elif len(self.__open_nodes) > 0:
            self.__current_node = self.__get_highest_priority_node(self.__open_nodes)
            logdebug(LOGGER, 'Selected an open node: %s', self.__current_node['host'])

        else:
            if self.__current_node is None:
                logwarn(LOGGER, 'Unexpected: No RabbitMQ node left to try, and there is no current one.')
                raise esgfpid.exceptions.ArgumentError('No RabbitMQ nodes were passed at all.')
            logwarn(LOGGER, 'No RabbitMQ node left to try! Leaving the last one: %s', self.__current_node['host'])

        self.__exchange_name = self.__current_node['exchange_name']

    def __get_highest_priority_node(self, dict_of_nodes):

        # Get highest priority:
        available_priorities = list(dict_of_nodes.keys())
        available_priorities.sort(key=natural_keys)
        current_priority = available_priorities.pop(0)
        list_of_priority_nodes = dict_of_nodes[current_priority]

        # Select one of them
        if len(list_of_priority_nodes)==1:
            nexthost = list_of_priority_nodes.pop()
            if len(list_of_priority_nodes)==0:
                dict_of_nodes.pop(current_priority)
            return nexthost
        else:
            nexthost = self.__select_and_remove_random_url_from_list(list_of_priority_nodes)
            return nexthost
        # TODO WHAT IF NONE LEFT???

    '''
    Return a pika.BasicProperties object needed for
    connecting to RabbitMQ.

    This returns always the same. Does not depend on node.

    :return: A  properties object (pika.BasicProperties).'''
    def get_properties_for_message_publications(self):
        return self.__properties

    '''
    Select and return a random URL from a list.
    This modifies the list and returns the URL!

    :param list_urls: The list of URLs to randomly select from.
    :return: Randomly selected URL.
    '''
    def __select_and_remove_random_url_from_list(self, list_urls):
        num_urls = len(list_urls)
        random_num = random.randint(0,num_urls-1)
        selected_url = list_urls[random_num]
        list_urls.remove(selected_url)
        return selected_url

    '''
    Return the current exchange name as string.

    :return: The current exchange name.
    '''
    def get_exchange_name(self):
        return self.__exchange_name

    '''
    Adapt the routing key (if necessary) to indicate 
    that a message comes from an untrusted node.

    The middle part (the RabbitMQ instruction part)
    of the key is adapted so it is visible that the
    node is an untrusted one.

    Note: The binding has to be done in the RabbitMQ exit 
    node (by the consumer).

    :return: The adapted routing key, in case the node
             is untrusted.
    '''
    def adapt_routing_key_for_untrusted(self, routing_key):

        # Message is published via an open node:
        if self.__current_node['is_open'] == True:
            if self.__has_trusted:
                return esgfpid.utils.adapt_routing_key_for_untrusted_fallback(routing_key)
            else:
                return esgfpid.utils.adapt_routing_key_for_untrusted(routing_key)

        # Message is published via a trusted node:
        elif self.__current_node['is_open'] == False:
            return routing_key
        else:
            logerror(LOGGER, 'Problem: Unsure whether the current node is open or not!')
            return esgfpid.utils.adapt_routing_key_for_untrusted_fallback(routing_key)

    '''
    Reset the list of available RabbitMQ instances to
    how it was before trying any.
    Once all RabbitMQ instances have been tried,
    the list is reset, so we can start over trying
    to connect.

    TODO: Where is this called?
    '''
    def reset_nodes(self):
        logdebug(LOGGER, 'Resetting hosts...')

        # Create a copy of a node list by rebuilding the ConnectionParameters with the values of the original nodes.
        # This is done to prevent performing a copy of the SSLContext in the pika SSLOption object.
        def _copy_node_list(node_list):
            list_copy = {}
            ignore_keys = ['params']
            for list_key, list_value in node_list.items():
                copy_value = []
                for node_info in list_value:
                    node_info_copy = dict((k, v) for k,v in node_info.items() if k not in ignore_keys)
                    copy_value.append(self.__complete_info_dict(node_info_copy, node_info['is_open']))
                list_copy[list_key] = copy_value
            return list_copy

        self.__trusted_nodes = _copy_node_list(self.__trusted_nodes_archive)
        self.__open_nodes = _copy_node_list(self.__open_nodes_archive)

        self.set_next_host()

    def _get_prio_stored_for_current(self):
        # Currently only used in unit test
        return self.__current_node['priority']

    def _get_prio_where_current_is_stored(self):
        # Currently only used in unit test

        if self.__current_node['is_open']:
            where_to_look = self.__open_nodes_archive
        else:
            where_to_look = self.__trusted_nodes_archive

        if not type(where_to_look) == type(dict()):
            raise ValueError('%s is not a dict!' % where_to_look)

        for prio, nodes in where_to_look.items():
            for candidate in nodes:
                if self.__compare_nodes(self.__current_node, candidate):
                    return prio

        raise ValueError('Node not found, so could not know currently active prio!')



import unittest
import mock
import logging
import copy
import os
import esgfpid.rabbit.asynchronous.thread_confirmer

LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())

UNCONFIRMED_TAGS = [1,2,3,4]
UNCONFIRMED_MESSAGES = {'1':'foo1', '2':'foo2', '3':'foo3', '4':'foo4'}

class ThreadConfirmerTestCase(unittest.TestCase):

    def setUp(self):
        LOGGER.info('######## Next test (%s) ##########', __name__)

    def tearDown(self):
        LOGGER.info('#############################')

    def make_confirmer(self):
        confirmer = esgfpid.rabbit.asynchronous.thread_confirmer.Confirmer()
        confirmer._Confirmer__unconfirmed_delivery_tags = copy.copy(UNCONFIRMED_TAGS)
        confirmer._Confirmer__unconfirmed_messages_dict = copy.copy(UNCONFIRMED_MESSAGES)
        return confirmer

    # Tests

    #
    # Acks
    #

    def test_single_ack_ok(self):

        # Preparation:
        confirmer = self.make_confirmer()
                
        # Make a fake ack:
        method_frame = mock.MagicMock()
        method_frame.method.delivery_tag = 2
        method_frame.method.multiple = False
        method_frame.method.NAME = 'foo.ack'

        # Run code to be tested:
        confirmer.on_delivery_confirmation(method_frame)

        # Check result (delivery tags)
        unconf = confirmer._Confirmer__unconfirmed_delivery_tags
        expected_unconf = [1,3,4]
        self.assertEqual(unconf, expected_unconf,
            'Unconfirmed delivery tags: %s, expected %s' % (unconf, expected_unconf))
        # Check result (messages)
        unconf_dict = confirmer._Confirmer__unconfirmed_messages_dict
        expected_unconf_dict = {'1':'foo1', '3':'foo3', '4':'foo4'}
        self.assertEqual(unconf_dict, expected_unconf_dict,
            'Unconfirmed messages: %s, expected %s' % (unconf_dict, expected_unconf_dict))


    def test_multiple_ack_ok(self):

        # Preparation:
        confirmer = self.make_confirmer()
        
        # Make a fake ack:
        method_frame = mock.MagicMock()
        method_frame.method.delivery_tag = 2
        method_frame.method.multiple = True
        method_frame.method.NAME = 'foo.ack'

        # Run code to be tested:
        confirmer.on_delivery_confirmation(method_frame)

        # Check result (delivery tags)
        unconf = confirmer._Confirmer__unconfirmed_delivery_tags
        expected_unconf = [3,4]
        self.assertEqual(unconf, expected_unconf,
            'Unconfirmed delivery tags: %s, expected %s' % (unconf, expected_unconf))
        # Check result (messages)
        unconf_dict = confirmer._Confirmer__unconfirmed_messages_dict
        expected_unconf_dict = {'3':'foo3', '4':'foo4'}
        self.assertEqual(unconf_dict, expected_unconf_dict,
            'Unconfirmed messages: %s, expected %s' % (unconf_dict, expected_unconf_dict))

    #
    # Nacks
    #

    def test_single_nack_ok(self):

        # Preparation:
        confirmer = self.make_confirmer()
                
        # Make a fake ack:
        method_frame = mock.Mock()
        method_frame.method.delivery_tag = 2
        method_frame.method.multiple = False
        method_frame.method.NAME = 'foo.nack'

        # Run code to be tested:
        confirmer.on_delivery_confirmation(method_frame)

        # Check result (delivery tags)
        unconf = confirmer._Confirmer__unconfirmed_delivery_tags
        expected_unconf = [1,3,4]
        self.assertEqual(unconf, expected_unconf,
            'Unconfirmed delivery tags: %s, expected %s' % (unconf, expected_unconf))
        # Check result (messages)
        unconf_dict = confirmer._Confirmer__unconfirmed_messages_dict
        expected_unconf_dict = {'1':'foo1', '3':'foo3', '4':'foo4'}
        self.assertEqual(unconf_dict, expected_unconf_dict,
            'Unconfirmed messages: %s, expected %s' % (unconf_dict, expected_unconf_dict))
        # Check result (nacked)
        nacked = confirmer._Confirmer__nacked_messages
        exp = ['foo2']
        self.assertEqual(nacked, exp,
            'Nacked messages: %s, expected %s' % (nacked, exp))

    def test_multiple_nack_ok(self):

        # Preparation:
        confirmer = self.make_confirmer()
        
        # Make a fake ack:
        method_frame = mock.Mock()
        method_frame.method.delivery_tag = 2
        method_frame.method.multiple = True
        method_frame.method.NAME = 'foo.nack'

        # Run code to be tested:
        confirmer.on_delivery_confirmation(method_frame)
        nacked_copy = confirmer.get_copy_of_nacked()

        # Check result (delivery tags)
        unconf = confirmer._Confirmer__unconfirmed_delivery_tags
        expected_unconf = [3,4]
        self.assertEqual(unconf, expected_unconf,
            'Unconfirmed delivery tags: %s, expected %s' % (unconf, expected_unconf))
        # Check result (messages)
        unconf_dict = confirmer._Confirmer__unconfirmed_messages_dict
        expected_unconf_dict = {'3':'foo3', '4':'foo4'}
        self.assertEqual(unconf_dict, expected_unconf_dict,
            'Unconfirmed messages: %s, expected %s' % (unconf_dict, expected_unconf_dict))
        # Check result (nacked)
        nacked = confirmer._Confirmer__nacked_messages
        exp = ['foo1', 'foo2']
        self.assertEqual(nacked, exp,
            'Nacked messages: %s, expected %s' % (nacked, exp))
        # Check result (nacked)
        exp = ['foo1', 'foo2']
        self.assertEqual(nacked_copy, exp,
            'Nacked messages: %s, expected %s' % (nacked_copy, exp))

    #
    # During publish
    #

    def test_putting_messages_ok(self):

        # Preparation:
        confirmer = self.make_confirmer()

        # Run code to be tested:
        confirmer.put_to_unconfirmed_delivery_tags(100)
        confirmer.put_to_unconfirmed_messages_dict(100, 'foo100')

        # Check result (delivery tags)
        unconf = confirmer._Confirmer__unconfirmed_delivery_tags
        expected_unconf = [1,2,3,4,100]
        self.assertEqual(unconf, expected_unconf,
            'Unconfirmed delivery tags: %s, expected %s' % (unconf, expected_unconf))
        # Check result (messages)
        unconf_dict = confirmer._Confirmer__unconfirmed_messages_dict
        expected_unconf_dict = {'1':'foo1', '2':'foo2', '3':'foo3', '4':'foo4', '100':'foo100'}
        self.assertEqual(unconf_dict, expected_unconf_dict,
            'Unconfirmed messages: %s, expected %s' % (unconf_dict, expected_unconf_dict))

    #
    # Getting leftovers
    #

    def test_getting_leftovers_ok(self):

        # Preparation:
        confirmer = self.make_confirmer()

        # Run code to be tested:
        mylist = confirmer.get_unconfirmed_messages_as_list_copy()
        num = confirmer.get_num_unconfirmed()

        # Check result (delivery tags)
        expected_messages = ['foo1', 'foo2', 'foo3', 'foo4']
        self.assertEqual(set(mylist), set(expected_messages),
            'Leftover messages: %s, expected %s' % (mylist, expected_messages))
        self.assertEqual(len(mylist), 4)
        self.assertEqual(num, 4)

    #
    # Reset
    #
 
    def test_reset_ok(self):
    
        # Preparation:
        confirmer = self.make_confirmer()
        
        # Run code to be tested:
        confirmer.reset_unconfirmed_messages_and_delivery_tags()
        unconf_retrieved = confirmer.get_copy_of_unconfirmed_tags()
    
        # Check result (delivery tags)
        unconf = confirmer._Confirmer__unconfirmed_delivery_tags
        self.assertEqual(unconf, [],
            'Unconfirmed delivery tags: %s, expected %s' % (unconf, []))
        self.assertEqual(unconf_retrieved, [],
            'Unconfirmed delivery tags: %s, expected %s' % (unconf_retrieved, []))

    #
    # Error
    #

    def test_neither_ack_nor_nack_error(self):

        # Preparation:
        confirmer = self.make_confirmer()
                
        # Make a fake ack:
        method_frame = mock.MagicMock()
        method_frame.method.delivery_tag = 2
        method_frame.method.multiple = False
        method_frame.method.NAME = 'foo.foo'

        # Run code to be tested:
        with self.assertRaises(esgfpid.rabbit.asynchronous.exceptions.UnknownServerResponse) as e:
            confirmer.on_delivery_confirmation(method_frame)
        self.assertIn('unknown type', repr(e.exception))

    def test_server_response_error_1(self):

        # Preparation:
        confirmer = self.make_confirmer()
                
        # Make a fake ack:
        method_frame = mock.MagicMock()
        method_frame.method.delivery_tag = 2
        method_frame.method.multiple = False
        method_frame.method.NAME = 'foo' # has no dot!

        # Run code to be tested:
        with self.assertRaises(esgfpid.rabbit.asynchronous.exceptions.UnknownServerResponse) as e:
            confirmer.on_delivery_confirmation(method_frame)
        self.assertIn('list index out of range', repr(e.exception))

    def test_server_response_error_2(self):

        # Preparation:
        confirmer = self.make_confirmer()
                
        # Make a fake ack:
        method_frame = mock.MagicMock()
        method_frame.method.delivery_tag = 2
        method_frame.method.multiple = False
        method_frame.method.NAME = None # is no string!

        # Run code to be tested:
        with self.assertRaises(esgfpid.rabbit.asynchronous.exceptions.UnknownServerResponse) as e:
            confirmer.on_delivery_confirmation(method_frame)
        self.assertIn('\\\'NoneType\\\' object has no attribute \\\'split\\\'', repr(e.exception))

 


import logging
import esgfpid.exceptions
from . import utils as solrutils

LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())

class FindVersionsOfSameDataset(object):

    def __init__(self, solr_interactor):
        self.__solr_interactor = solr_interactor
        self.__error_messages = None

    def __reset_error_messages(self):
        self.__error_messages = []

    # General methods:

    def retrieve_dataset_handles_or_version_numbers_of_all_versions(self, drs_id, prefix):
        '''
        :return: Dict. Should never return None.
        :raise: SolrSwitchedOff
        :raise SolrError: If ...
        '''
        self.__reset_error_messages()
        response_json = self.__ask_solr_for_handles_or_version_numbers_of_all_versions(drs_id)
        result_dict = self.__parse_result_handles_or_version_numbers_of_all_versions(response_json, prefix)
        return result_dict

    # Querying solr for handles and/or version numbers of all versions of a dataset (same drs_id):

    def __ask_solr_for_handles_or_version_numbers_of_all_versions(self, drs_id):
        LOGGER.debug('Asking solr for dataset handles or version numbers of all dataset versions with the drs_id "%s".', drs_id)
        query = self.__make_query_for_handles_or_version_numbers_of_all_versions(drs_id)
        LOGGER.debug('Query: %s', query)
        response_json = self.__solr_interactor.send_query(query) # can raise SolrError or SolrSwitchedOff, but can't be None
        return response_json

    def __make_query_for_handles_or_version_numbers_of_all_versions(self, drs_id):
        query_dict = self.__solr_interactor.make_solr_base_query()
        query_dict['type'] = 'Dataset'
        query_dict['facets'] = 'pid,version'
        query_dict['drs_id'] = drs_id
        return query_dict

    def __parse_result_handles_or_version_numbers_of_all_versions(self, response_json, prefix):

        # Prepare result object:
        result_dict = {}
        result_dict['dataset_handles'] = None
        result_dict['version_numbers'] = None

        # Get handles, if there is any:
        result_dict['dataset_handles'] = self.__get_handles_if_any(response_json, prefix)

        # Otherwise, get version numbers, if there is any:
        result_dict['version_numbers'] = self.__get_version_numbers_if_any(response_json)

        if (result_dict['dataset_handles'] is None and
            result_dict['version_numbers'] is None):

            msg = 'Found neither version numbers, nor handles. Errors: %s' % '; '.join(self.__error_messages)
            LOGGER.debug(msg)
            raise esgfpid.exceptions.SolrResponseError(msg)

        return result_dict

    def __get_handles_if_any(self, response_json, prefix):
        try:
            dataset_handles = solrutils.extract_dataset_handles_from_response_json(response_json, prefix)
            if len(dataset_handles) > 0:
                LOGGER.debug('Found dataset handles: %s', dataset_handles)
            return dataset_handles

        except esgfpid.exceptions.SolrResponseError as e:
            self.__error_messages.append(repr(e))

    def __get_version_numbers_if_any(self, response_json):
        try:
            version_numbers = solrutils.extract_dataset_version_numbers_from_response_json(response_json)
            if len(version_numbers) > 0:
                LOGGER.debug('Found version numbers, but no handles: %s', version_numbers)
            return version_numbers

        except esgfpid.exceptions.SolrResponseError as e:
            self.__error_messages.append(repr(e))



import unittest
import mock
import logging
import esgfpid
import sys
if sys.version_info.major >= 3:
    import importlib
    def reload(module):
        return importlib.reload(module)

LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())


class LoggerMock(object):

    def __init__(self):
        self.info_messages = []
        self.debug_messages = []
        self.error_messages = []
        self.warn_messages = []

    def info(self, msg, *args, **kwargs):
        msgf = msg % args
        self.info_messages.append(msgf)

    def debug(self, msg, *args, **kwargs):
        logging.warn('DEBUG! '+str(msg))
        msgf = msg % args
        self.debug_messages.append(msgf)

    def warn(self, msg, *args, **kwargs):
        msgf = msg % args
        self.warn_messages.append(msgf)

    def error(self, msg, *args, **kwargs):
        msgf = msg % args
        self.error_messages.append(msgf)

class UtilsLoggingTestCase(unittest.TestCase):

    def setUp(self):
        LOGGER.info('######## Next test (%s) ##########', __name__)

    def tearDown(self):
        LOGGER.info('#############################')

    #
    # Test whether args are passed ok!
    #

    @mock.patch('esgfpid.defaults')
    def test_log_info_pass_args_kwargs(self, defaults_patch):
        ''' Test if info messages stay info when flag is not set! '''

        # Test variables
        logger = LoggerMock()

        # Prepare flags
        defaults_patch.LOG_INFO_TO_DEBUG = False # <-- This is tested
        defaults_patch.LOG_TRACE_TO_DEBUG = 999 # irrelevant here
        reload(esgfpid.utils)

        # Run code to be tested:
        esgfpid.utils.loginfo(logger, 'barbaz %s %i', 'a', 1, show=True)

        # Check results:
        expected_info = 'barbaz a 1'
        self.assertIn(expected_info, logger.info_messages, 'Received messages: %s' % logger.info_messages)

    #
    # Test whether correct levels are reached
    #

    @mock.patch('esgfpid.defaults')
    def test_log_info_normal(self, defaults_patch):
        ''' Test if info messages stay info when flag is not set! '''

        # Test variables
        logger = LoggerMock()

        # Prepare flags
        defaults_patch.LOG_INFO_TO_DEBUG = False # <-- This is tested
        defaults_patch.LOG_TRACE_TO_DEBUG = 999 # irrelevant here
        reload(esgfpid.utils)

        # Run code to be tested:
        esgfpid.utils.loginfo(logger, 'barbaz')

        # Check results:
        self.assertIn('barbaz', logger.info_messages, 'Nope')

    @mock.patch('esgfpid.defaults')
    def test_log_info_to_debug(self, defaults_patch):
        ''' Test if info messages are made debug when flag is set! '''

        # Test variables
        logger = LoggerMock()

        # Prepare flags
        defaults_patch.LOG_INFO_TO_DEBUG = True # <-- This is tested
        defaults_patch.LOG_TRACE_TO_DEBUG = 999 # irrelevant here
        reload(esgfpid.utils)

        # Run code to be tested:
        esgfpid.utils.loginfo(logger, 'barbaz')

        # Check results:
        self.assertIn('barbaz', logger.debug_messages, 'Nope')

    @mock.patch('esgfpid.defaults')
    def test_log_debug_normal(self, defaults_patch):
        ''' Test normal debug messages! '''

        # Test variables
        logger = LoggerMock()

        # Prepare flags
        defaults_patch.LOG_INFO_TO_DEBUG = 999 # irrelevant here
        defaults_patch.LOG_DEBUG_TO_INFO = False
        defaults_patch.LOG_TRACE_TO_DEBUG = 999 # irrelevant here
        reload(esgfpid.utils)

        # Run code to be tested:
        esgfpid.utils.logdebug(logger, 'foofoo')

        # Check results:
        self.assertEqual(logger.info_messages, [], 'Received info: %s (should be an an empty list).' % (logger.info_messages))
        self.assertIn('foofoo', logger.debug_messages, 'Missing "foofoo" in: '+str(logger.debug_messages))

    @mock.patch('esgfpid.defaults')
    def test_log_trace_normal_ignored(self, defaults_patch):
        ''' Test normal trace messages! '''

        # Test variables
        logger = LoggerMock()

        # Prepare flags
        defaults_patch.LOG_INFO_TO_DEBUG = 999 # irrelevant here
        defaults_patch.LOG_TRACE_TO_DEBUG = False #  this is relevant!
        reload(esgfpid.utils)

        # Run code to be tested:
        esgfpid.utils.logtrace(logger, 'superdetail')

        # Check results:
        self.assertEqual(logger.info_messages, [], 'Received info: %s (should be an an empty list).' % (logger.info_messages))
        self.assertEqual(logger.debug_messages, [], 'Received debug: %s (should be an empty list).' % (logger.debug_messages))

    @mock.patch('esgfpid.defaults')
    def test_log_trace_to_debug(self, defaults_patch):
        ''' Test trace messages that are shown as DEBUG '''

        # Test variables
        logger = LoggerMock()

        # Prepare flags
        defaults_patch.LOG_INFO_TO_DEBUG = 999 # irrelevant here
        defaults_patch.LOG_TRACE_TO_DEBUG = True #  this is relevant!
        defaults_patch.LOG_DEBUG_TO_INFO = False
        reload(esgfpid.utils)

        # Run code to be tested:
        esgfpid.utils.logtrace(logger, 'superdetail')

        # Check results:
        expected_debug = '[trace] superdetail'
        self.assertTrue(esgfpid.defaults.LOG_TRACE_TO_DEBUG)
        self.assertIn(expected_debug, logger.debug_messages, 'Received debug: %s (should be %s).' % (logger.debug_messages, expected_debug))
        self.assertEqual(logger.info_messages, [], 'Received info: %s (should be empty list).' % (logger.info_messages))


    @mock.patch('esgfpid.defaults')
    def test_log_warn_normal(self, defaults_patch):
        ''' Test if info messages stay info when flag is not set! '''

        # Test variables
        logger = LoggerMock()

        # Prepare flags
        defaults_patch.LOG_INFO_TO_DEBUG = 999 # irrelevant here
        defaults_patch.LOG_TRACE_TO_DEBUG = 999 # irrelevant here
        reload(esgfpid.utils)

        # Run code to be tested:
        esgfpid.utils.logwarn(logger, 'danger')

        # Check results:
        self.assertIn('danger', logger.warn_messages, 'Warn messages: %s' % logger.warn_messages)

    @mock.patch('esgfpid.defaults')
    def test_log_error_normal(self, defaults_patch):
        ''' Test if info messages stay info when flag is not set! '''

        # Test variables
        logger = LoggerMock()

        # Prepare flags
        defaults_patch.LOG_INFO_TO_DEBUG = 999 # irrelevant here
        defaults_patch.LOG_TRACE_TO_DEBUG = 999 # irrelevant here
        reload(esgfpid.utils)

        # Run code to be tested:
        esgfpid.utils.logerror(logger, 'danger')

        # Check results:
        self.assertIn('danger', logger.error_messages, 'Error messages: %s' % logger.error_messages)


    #
    # Logging every xth message
    #

    @mock.patch('esgfpid.defaults')
    def test_log_every_x_times_normal_debug_ok(self, defaults_patch):
        '''
        Check ifmessage is logged every 10th time, and the first time.
        Messages should be printed as debug.
        '''

        # Test variables
        logger = LoggerMock()
        counter = 0
        x = 10
        msg = 'Foobar'

        # Prepare flags
        defaults_patch.LOG_INFO_TO_DEBUG = 999 # irrelevant here
        defaults_patch.LOG_TRACE_TO_DEBUG = 999 # irrelevant here
        defaults_patch.LOG_DEBUG_TO_INFO = False
        reload(esgfpid.utils)

        # Run code to be tested:
        for i in range(35):
            counter += 1
            esgfpid.utils.log_every_x_times(logger, counter, x, (msg+str(counter)))

        # Check result
        received_messages = ', '.join(logger.debug_messages)
        #expected_messages = 'Foobar1 (counter 1), Foobar10 (counter 10), Foobar20 (counter 20), Foobar30 (counter 30)'
        expected_messages = 'Foobar1, Foobar10, Foobar20, Foobar30'
        self.assertEqual([],logger.info_messages,'Received info messages: "%s" (should be empty list).' % logger.info_messages)
        self.assertEqual(expected_messages, received_messages, 'Received messages: %s (should be %s).' % (received_messages,expected_messages))


    def test_make_logsafe_list_normal(self):
        input_list = [{'a':'aaa', 'b':'bbb', 'password':'einekleinenachtmusik'},
                      {'a':'111', 'b':'222', 'c':'333'}]
        output = esgfpid.utils.make_logsafe(input_list)
        self.assertTrue('einekleinenachtmusik' not in str(output))

    def test_make_logsafe_list_several(self):
        input_list = [{'a':'aaa', 'b':'bbb', 'password':'einekleinenachtmusik'},
                      {'a':'111', 'b':'222', 'password':'unpetitemusique'}]
        output = esgfpid.utils.make_logsafe(input_list)
        self.assertTrue('einekleinenachtmusik' not in str(output))
        self.assertTrue('unpetitemusique' not in str(output))
    
    def test_make_logsafe_list_none(self):
        input_list = [{'a':'aaa', 'b':'bbb', 'c':'einekleinenachtmusik'},
                      {'a':'111', 'b':'222', 'c':'unpetitemusique'}]
        output = esgfpid.utils.make_logsafe(input_list)
        self.assertEqual(input_list, output)

    def test_make_logsafe_list_differentword(self):
        input_list = [{'a':'aaa', 'b':'bbb', 'c':'einekleinenachtmusik'},
                      {'a':'111', 'b':'222', 'motdepasse':'unpetitemusique'}]
        output = esgfpid.utils.make_logsafe(input_list, 'motdepasse')
        self.assertTrue('unpetitemusique' not in str(output))

    def test_make_logsafe_dict_normal(self):
        input_dict = {'a':'aaa', 'b':'bbb', 'password':'einekleinenachtmusik'}
        output = esgfpid.utils.make_logsafe(input_dict)
        self.assertTrue('einekleinenachtmusik' not in str(output))

    def test_make_logsafe_dict_several(self):
        input_dict = {'a':'aaa', 'password':'maismusica', 'password':'einekleinenachtmusik'}
        output = esgfpid.utils.make_logsafe(input_dict)
        self.assertTrue('einekleinenachtmusik' not in str(output))
        self.assertTrue('maismusica' not in str(output))

    def test_make_logsafe_dict_differentword(self):
        input_dict = {'a':'aaa', 'b':'bbb', 'motdepasse':'einekleinenachtmusik'}
        output = esgfpid.utils.esgfpid.utils.make_logsafe(input_dict, 'motdepasse')
        self.assertTrue('einekleinenachtmusik' not in str(output))



import logging
import esgfpid.defaults as defaults
from esgfpid.utils import loginfo, logdebug, logtrace, logerror, logwarn, log_every_x_times
from esgfpid.utils import get_now_utc_as_formatted_string as get_now_utc_as_formatted_string
import time

LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())

class ShutDowner(object):

    def __init__(self, thread, statemachine):
        self.thread = thread
        self.statemachine = statemachine

        '''
        To count how many times we've iterated during
        the iterative, recursive check for pending messages.
        '''
        self.__close_decision_iterations = 0

        '''
        To see, in case of a reconnect, if the module was in
        the process of gently finishing, so we can add that
        to the new ioloop.
        Otherwise, the gently-closing is lost during a reconnect.
        '''
        self.__is_in_process_of_gently_closing = False

    ####################
    ### Force finish ###
    ####################

    def force_finish(self):
        try:
            return self.__force_finish('Forced finish from outside the thread.')
        except Exception as e:
            logwarn(LOGGER, 'Error in shutter.force_finish(): %s: %s', e.__class__.__name__, repr(e))
            raise e

    #####################
    ### Gentle finish ###
    #####################

    def finish_gently(self):
        try:
            return self.__finish_gently()
        except Exception as e:
            logwarn(LOGGER, 'Error in shutter.finish_gently(): %s: %s', e.__class__.__name__, repr(e))
            raise e

    def __finish_gently(self):
        # Called directly from outside the thread!

        # No more messages can arrive from publisher (because
        # the main thread blocks), but publishes/confirms are still
        # accepted.

        # Inform user
        if self.__are_any_messages_pending():
            wait_seconds = defaults.RABBIT_ASYN_FINISH_WAIT_SECONDS
            max_waits = defaults.RABBIT_ASYN_FINISH_MAX_TRIES
            loginfo(LOGGER, 'Preparing to close PID module. Some messages are pending. Maximum waiting time: %i seconds. (%s)', wait_seconds*max_waits, get_now_utc_as_formatted_string())
        else:
            loginfo(LOGGER, 'Closing PID module. No pending messages. (%s)', get_now_utc_as_formatted_string())

        # Go through decision tree (close or wait for pending messages)
        self.__close_decision_iterations = 1
        self.__is_in_process_of_gently_closing = True
        self.recursive_decision_about_closing()
        # This iteratively checks if all messages are published+confirmed.
        # If not, it waits and then rechecks, up to a maximum number of iterations.
        # The main thread waits for this by using a threading.Event.

    ''' Called by builder (via thread), so close-events are not lost if a new ioloop is started.'''
    def continue_gently_closing_if_applicable(self):
        if self.__is_in_process_of_gently_closing:
            logdebug(LOGGER, 'Continue gentle shutdown even after reconnect (iteration %i)...', self.__close_decision_iterations)
            if self.thread._connection is not None:
                wait_seconds = defaults.RABBIT_ASYN_FINISH_WAIT_SECONDS
                self.thread._connection.ioloop.call_later(wait_seconds, self.recursive_decision_about_closing)
            else:
                logerror(LOGGER, 'Connection was None when trying to wait for pending messages (after reconnect). Synchronization error between threads!')

    def recursive_decision_about_closing(self):
        logdebug(LOGGER, 'Gentle finish (iteration %i): Deciding about whether we can close the thread or not...', self.__close_decision_iterations)
        iteration = self.__close_decision_iterations
        if self.__are_any_messages_pending():
            self.__inform_about_pending_messages()
            self.__decide_what_to_do_about_pending_messages(iteration)
        else:
            self.__close_because_all_done(iteration)

    def __decide_what_to_do_about_pending_messages(self, iteration):
        logdebug(LOGGER, 'Gentle finish (iteration %i): Decide what to do about the pending messages...', self.__close_decision_iterations)
        if self.__have_we_waited_enough_now(iteration):
            self.__close_because_waited_long_enough()
        elif self.__module_is_not_progressing_anymore():
            self.__close_because_no_point_in_waiting()
        else:
            # Make sure the messages can be sent, in case some events
            # were lost during reconnecting or something...
            num_unpub = self.thread.get_num_unpublished()
            logdebug(LOGGER, 'Triggering %i publish events...' % num_unpub)
            for i in range(int(1.1*num_unpub)):
                self.thread.add_event_publish_message()
            # Now wait some more...
            self.__wait_some_more_and_redecide(iteration)

    # Decision rules:

    def __have_we_waited_enough_now(self, iteration):
        logdebug(LOGGER, 'Gentle finish (iteration %i): Check if the rabbit thread has waited long enough...', self.__close_decision_iterations)

        wait_seconds = defaults.RABBIT_ASYN_FINISH_WAIT_SECONDS
        max_waits = defaults.RABBIT_ASYN_FINISH_MAX_TRIES

        tried = iteration
        waited = iteration-1

        # Logging:
        logdebug(LOGGER, 'Gentle finish (iteration %i): At this point we have tried %i times and waited %i/%i times (%i seconds)', self.__close_decision_iterations, tried, waited, max_waits, waited*wait_seconds)
        log_every_x_seconds = 2
        if ((waited*wait_seconds)%log_every_x_seconds==0 or waited>=max_waits):
            msg = self.__get_string_about_pending_messages()
            loginfo(LOGGER, 'Still pending: %s messages... (waited %.1f/%.1f seconds)', msg, waited*wait_seconds, max_waits*wait_seconds)
        
        # Return:
        if waited >= max_waits:
            logdebug(LOGGER, 'Gentle finish (iteration %i): The rabbit thread has waited long enough for pending messages at close down.', self.__close_decision_iterations)
            return True
        logdebug(LOGGER, 'Gentle finish (iteration %i): We should wait a little more for pending messages.', self.__close_decision_iterations)
        return False

    def __are_any_messages_pending(self):
        logdebug(LOGGER, 'Gentle finish (iteration %i): Checking for any pending messages...', self.__close_decision_iterations)
        sent_done = self.__check_all_were_sent()
        confirmed_done = self.__check_all_were_confirmed()
        if sent_done and confirmed_done:
            logdebug(LOGGER, 'Gentle finish (iteration %i): No more pending messages.', self.__close_decision_iterations)
            return False # none pending
        logdebug(LOGGER, 'Gentle finish (iteration %i): Some pending messages left.', self.__close_decision_iterations)
        return True # some are pending

    def __check_all_were_sent(self):
        approx_number_of_messages_in_queue = self.thread.get_num_unpublished()
        if approx_number_of_messages_in_queue == 0:
            return True
        else:
            return False

    def __check_all_were_confirmed(self):
        if self.thread.get_num_unconfirmed() == 0:
            return True
        else:
            return False

    def __module_is_not_progressing_anymore(self):
        if self.statemachine.is_PERMANENTLY_UNAVAILABLE() or self.statemachine.is_FORCE_FINISHED():
            logdebug(LOGGER, 'Gentle finish (iteration %i): The rabbit thread is not active anymore, so we might as well close it.', self.__close_decision_iterations)
            return True
        return False

    def __wait_some_more_and_redecide(self, iteration):
        wait_seconds = defaults.RABBIT_ASYN_FINISH_WAIT_SECONDS
        logdebug(LOGGER, 'Gentle finish (iteration %i): Waiting some more for pending messages...', self.__close_decision_iterations)
        # Instead of time.sleep(), add an event to the thread's ioloop
        self.__close_decision_iterations += 1
        if self.thread._connection is not None:
            self.thread._connection.ioloop.call_later(wait_seconds, self.recursive_decision_about_closing)
            self.__is_in_process_of_gently_closing = True
            # Problem: If a reconnect occurs after this, this event will be lost.
            # I cannot retrieve if from the ioloop and pass it to the new one.
            # So, during a reconnection, we check if gently-finish was running,
            # and add a new timeout to the new ioloop, using
            # "continue_gently_closing_if_applicable()".
            # This may mess up the amount of time the gently-finish takes, though.
            # TODO Maybe one day it is possible to transfer events from one ioloop to
            # another?
        else:
            logerror(LOGGER, 'Connection was None when trying to wait for pending messages. Synchronization error between threads!')

    
    def __tell_publisher_to_stop_waiting_for_gentle_finish(self):
        logdebug(LOGGER, 'Main thread does not need to wait anymore. (%s).', get_now_utc_as_formatted_string())

        # This avoids that the last iteration is redone
        # and redone upon reconnection, as after reconnection,
        # if this is True, the algorithm is entered again.
        self.__is_in_process_of_gently_closing = False

        # This releases the event that blocks the main thread
        # until the gentle finish is done.
        self.thread.tell_publisher_to_stop_waiting_for_gentle_finish()

    def __close_because_all_done(self, iteration):
        logdebug(LOGGER, 'Gentle finish (iteration %i): All messages sent and confirmed in %ith try (waited and rechecked %i times).', self.__close_decision_iterations, iteration, iteration-1)
        loginfo(LOGGER, 'All messages sent and confirmed. Closing.')
        self.__normal_finish()
        self.__tell_publisher_to_stop_waiting_for_gentle_finish()

    def __close_because_no_point_in_waiting(self):

        # Logging, depending on why we closed...
        logdebug(LOGGER, 'Gentle finish (iteration %i): Closing, as there is no point in waiting any longer.', self.__close_decision_iterations)
        if self.statemachine.get_detail_closed_by_publisher():
            logwarn(LOGGER, 'Not waiting for pending messages: No connection to server (previously closed by user).')
        elif self.statemachine.detail_could_not_connect:
            logwarn(LOGGER, 'Not waiting for pending messages: No connection to server (unable to connect).')
        else:
            logwarn(LOGGER, 'Not waiting for pending messages: No connection to server (unsure why).')

        # Actual close
        self.__force_finish('Force finish as we are not sending the messages anyway.')
        self.__tell_publisher_to_stop_waiting_for_gentle_finish()

    def __close_because_waited_long_enough(self):
        logdebug(LOGGER, 'We have waited long enough. Now closing by force.')
        self.__force_finish('Force finish as normal waiting period in normal finish is over.')
        self.__tell_publisher_to_stop_waiting_for_gentle_finish()

    def __inform_about_pending_messages(self):
        msg = self.__get_string_about_pending_messages()
        if msg is not None:
            logdebug(LOGGER, 'Pending messages: %s.', msg)

    def __get_string_about_pending_messages(self):
        unsent = self.thread.get_num_unpublished()
        unconfirmed = self.thread.get_num_unconfirmed()
        if unsent + unconfirmed > 0:
            msg = '%i (%i unsent, %i unconfirmed)' % ((unsent+unconfirmed), unsent, unconfirmed)
            return msg
        else:
            return None

    #################
    ### Finishing ###
    #################

    def __force_finish(self, msg):
        logdebug(LOGGER, 'Force finishing, reason: %s.', msg)
        self.statemachine.set_to_force_finished()
        reply_code = self.thread.ERROR_CODE_CONNECTION_CLOSED_BY_USER
        reply_text = msg+' '+self.thread.ERROR_TEXT_CONNECTION_FORCE_CLOSED
        self.__close_down(reply_code, reply_text)
        self.__inform_about_state_at_shutdown()

    def __normal_finish(self, msg='Normal finish.'):
        reply_code = self.thread.ERROR_CODE_CONNECTION_CLOSED_BY_USER
        reply_text = msg+' '+self.thread.ERROR_TEXT_CONNECTION_NORMAL_SHUTDOWN
        self.__close_down(reply_code, reply_text)
        self.__inform_about_state_at_shutdown()

    def safety_finish(self, msg='Safety finish'):
        reply_code = self.thread.ERROR_CODE_CONNECTION_CLOSED_BY_USER
        reply_text = msg+' '+self.thread.ERROR_TEXT_CONNECTION_NORMAL_SHUTDOWN
        self.__close_down(reply_code, reply_text)

    def __close_down(self, reply_code, reply_text):
        # Important, so the connection does not get reopened by 
        # the on_connection_closed callback.
        # This should only be called by one of the finish methods.

        # Make sure the main thread does not continue blocking
        # as it believes that we're still looking for pending messages:
        self.__tell_publisher_to_stop_waiting_for_gentle_finish()

        # Change the state of the state machine:
        self.statemachine.set_to_permanently_unavailable()
        self.statemachine.set_detail_closed_by_publisher()

        # Close connection
        try:
            if self.thread._connection is not None:

                if self.thread._connection.is_closed or self.thread._connection.is_closing:
                    logdebug(LOGGER, 'Connection is closed or closing.')
                    # If connection is already closed, the on_connection_close is not
                    # called, so the ioloop continues, possibly waiting for reconnect.
                    # So we need to prevent reconnects or other events. As long
                    # as ioloop runs, thread cannot be finished/joined.
                    self.thread.make_permanently_closed_by_user()
                elif self.thread._connection.is_open:
                    logdebug(LOGGER, 'Connection is open. Closing now. This will trigger the RabbitMQ callbacks.')
                    self.thread._connection.close(reply_code=reply_code, reply_text=reply_text)
                    # "If there are any open channels, it will attempt to close them prior to fully disconnecting." (pika docs)
            else:
                logerror(LOGGER, 'Connection was None when trying to close. Synchronization error between threads!')

        except AttributeError as e:
            logdebug(LOGGER, 'AttributeError from pika during connection closedown (%s: %s)', e.__class__.__name__, repr(e))

    def __inform_about_state_at_shutdown(self):
        unsent = self.thread.get_num_unpublished()
        unconfirmed = self.thread.get_num_unconfirmed()
        if unsent + unconfirmed > 0:
            logwarn(LOGGER, 
                'At close down: %i pending messages (%i unpublished messages, %i unconfirmed messages).',
                (unsent+unconfirmed), unsent, unconfirmed)
        else:
            loginfo(LOGGER, 'After close down: All messages were published and confirmed.')



import logging
import datetime
import json
import pika
import time
import esgfpid.utils
import esgfpid.defaults as defaults
from esgfpid.utils import loginfo, logdebug, logtrace, logerror, logwarn
from esgfpid.exceptions import MessageNotDeliveredException
from esgfpid.utils import get_now_utc_as_formatted_string as get_now_utc_as_formatted_string
from .. import rabbitutils as rabbitutils
from ..exceptions import PIDServerException


LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())

class SynchronousRabbitConnector(object):

    '''
    Constructor for the sychronous rabbit connection module.

    This does not open a connection yet.

    :param node_manager: NodeManager object that contains 
        the info about all the available RabbitMQ instances,
        their credentials, their priorities.

    '''
    def __init__(self, nodemanager):

        loginfo(LOGGER, 'Init of SynchronousRabbitConnector!!! Bla')

        '''
        NodeManager provices info about all
        hosts.
        '''
        self.__nodemanager = nodemanager

        '''
        Props for basic_publish. Does not
        depend on host, so store it once for all.
        '''
        self.__props = self.__nodemanager.get_properties_for_message_publications()

        '''
        To count how many times we have tried to reconnect the set of
        RabbitMQ hosts.
        '''
        self.__reconnect_counter = 0

        '''
        To see how many times we should try reconnecting to the set 
        of RabbitMQ hosts. Note that if there is 3 hosts, and we try 2
        times, this means 6 connection tries in total.
        '''
        self.__max_reconnection_tries = defaults.RABBIT_RECONNECTION_MAX_TRIES

        '''
        How many seconds to wait before reconnecting after having tried
        all hosts. (There is no waiting time trying to connect to a different
        host after one fails).
        '''
        self.__wait_seconds_before_reconnect = defaults.RABBIT_RECONNECTION_SECONDS

        '''
        To see how much time it takes to connect. Once a connection is
        established or failed, we print the time delta to logs.
        '''
        self.__start_connect_time = None

        '''
        If the messages should not be published to the exchange that
        was passed from the publisher in config, but to a fallback 
        solution, this will be set:
        '''
        self.__fallback_exchange = None

        ''' Set of all tried hosts, for logging. '''
        self.__all_hosts_that_were_tried = set()



        # Defaults:
        self.__mandatory_flag = esgfpid.defaults.RABBIT_MANDATORY_DELIVERY
        self.__max_tries = esgfpid.defaults.RABBIT_SYN_MESSAGE_MAX_TRIES
        self.__timeout_milliseconds = esgfpid.defaults.RABBIT_SYN_MESSAGE_TIMEOUT_MILLISEC
        self.__candidate_fallback_exchange_name = defaults.RABBIT_FALLBACK_EXCHANGE_NAME

        # Other settings:
        self.__channel = None
        self.__connection = None
        self.__communication_established = False
        self.__connection_last_process_event_call = 0
        self.__error_messages_during_init = []


    ########################################
    ### This actually connects to rabbit ###
    ########################################

    '''
    Opens a connection and channel to RabbitMQ.

    It tries connecting to all available hosts until
    it succeeds, or until the maximum number of
    attempts per host is reached.

    The sequence of the hosts is determined
    by the NodeManager.

    The maximum number of connection attempts is
    determined in the esgfpid.defaults.

    '''
    def open_rabbit_connection(self):    

        continue_connecting = True
        while continue_connecting:

            success = self.__try_connecting_to_next()

            if success:
                continue_connecting = False

            else:

                # Log failure:
                oldhost = self.__nodemanager.get_connection_parameters().host
                time_passed = datetime.datetime.now() - self.__start_connect_time
                loginfo(LOGGER, 'Failed connection to RabbitMQ at %s after %s seconds.', oldhost, time_passed.total_seconds())

                # If there is alternative URLs, try one of them:
                if self.__nodemanager.has_more_urls():
                    logdebug(LOGGER, 'Connection failure: %s fallback URLs left to try.', self.__nodemanager.get_num_left_urls())
                    self.__nodemanager.set_next_host()
                    newhost = self.__nodemanager.get_connection_parameters().host
                    loginfo(LOGGER, 'Next connection attempt (now) %s.', newhost)

                # If there is no URLs, reset the node manager to
                # start at the first nodes again...
                else:
                    self.__reconnect_counter += 1;
                    if self.__reconnect_counter <= self.__max_reconnection_tries:
                        reopen_seconds = self.__wait_seconds_before_reconnect
                        logdebug(LOGGER, 'Connection failure: Failed connecting to all hosts. Waiting %s seconds and starting over.', reopen_seconds)
                        self.__nodemanager.reset_nodes()
                        newhost = self.__nodemanager.get_connection_parameters().host
                        loginfo(LOGGER, 'Next connection attempt (in %s seconds) to %s.', reopen_seconds, newhost)
                        time.sleep(reopen_seconds)

                    # Give up after so many tries...
                    else:
                        continue_connecting = False
                        errormsg = ('Permanently failed to connect to RabbitMQ. Tried all hosts %s %s times. Giving up. No PID requests will be sent.' % (list(self.__all_hosts_that_were_tried) ,self.__max_reconnection_tries))
                        logerror(LOGGER, errormsg)
                        collected_errors = ' - '.join(self.__error_messages_during_init)
                        logwarn(LOGGER, 'No connection possible. Errors: %s' % collected_errors)
                        raise PIDServerException(errormsg)

    '''
    Opens and stores connection and channel to the next
    RabbitMQ host. The order of the hosts is determined
    by the NodeManager object.

    :returns: True if successful. False otherwise.
    '''
    def __try_connecting_to_next(self):
        success = False

        # Open connection
        params = self.__nodemanager.get_connection_parameters()
        connection = self.__setup_rabbit_connection(params)

        # If connection is ok, open channel
        if connection is not None and connection.is_open:
            self.__connection = connection
            LOGGER.debug('Succeeded opening connection to '+str(params.host))
            channel = self.__setup_rabbit_channel()

            # If channel ok, success, yeah!
            if channel is not None and channel.is_open:
                LOGGER.debug('Succeeded opening channel.')
                self.__channel = channel
                self.__channel.confirm_delivery()
                self.__communication_established = True
                success = True
            else:
                LOGGER.debug('Could not open channel.')

        return success

    '''
    Opens the Connection to RabbitMQ instance and returns it.

    :param params: Parameters for connection to RabbitMQ, of
        type pika.ConnectionParameters.
    :returns: The connection, if it was built successfully. Returns
        None if there was an expected error that should be dealt 
        with by trying the next host.
    :raises: Any unexpected exception.
    '''
    def __setup_rabbit_connection(self, params):
        LOGGER.debug('Setting up the connection with the RabbitMQ.')
        self.__start_connect_time = datetime.datetime.now()
        self.__all_hosts_that_were_tried.add(params.host)
        
        try:
            time_now = get_now_utc_as_formatted_string()
            logdebug(LOGGER, 'Connecting to RabbitMQ at %s... (%s)', params.host, time_now)

            # Make connection
            conn = self.__make_connection(params)

            # Log success
            time_now = get_now_utc_as_formatted_string()
            time_passed = datetime.datetime.now() - self.__start_connect_time
            time_seconds = time_passed.total_seconds()
            loginfo(LOGGER, 'Connection to RabbitMQ at %s opened after %i seconds... (%s)', params.host, time_seconds, time_now)
            return conn

        except pika.exceptions.ProbableAuthenticationError as e:
            time_passed = datetime.datetime.now() - self.__start_connect_time
            time_seconds = time_passed.total_seconds()
            error_name = e.__class__.__name__
            logerror(LOGGER, 'Caught Authentication Exception after %s seconds during connection ("%s").', time_seconds, error_name)
            msg = ('Problem setting up the rabbit with username "%s" and password "%s" at url %s.' %
                    (params.credentials.username,
                     params.credentials.password,
                     params.host))
            LOGGER.error(msg)
            self.__error_messages_during_init.append(msg)
            return None

        except pika.exceptions.AMQPConnectionError as e:
            time_passed = datetime.datetime.now() - self.__start_connect_time
            time_seconds = time_passed.total_seconds()
            error_name = e.__class__.__name__
            logerror(LOGGER, 'Caught AMQPConnectionError Exception after %s seconds during connection ("%s").', time_seconds, error_name)
            msg = ('Problem setting up the rabbit connection to %s.' % params.host)
            self.__error_messages_during_init.append(msg)
            return None

        except Exception as e:
            time_passed = datetime.datetime.now() - self.__start_connect_time
            time_seconds = time_passed.total_seconds()
            error_name = e.__class__.__name__
            logerror(LOGGER, 'Error ("%s") during connection to %s, after %s seconds.',
                error_name, params.host, time_seconds)
            msg = ('Unexpected problem setting up the rabbit connection to %s (%s)' % (params.host, error_name))
            self.__error_messages_during_init.append(msg)
            raise e

    '''
    Little wrapper around pika's connection opening
    method, so we can mock this during unit tests.
    '''
    def __make_connection(self, params): # this is easy to mock during unit tests        
        return pika.BlockingConnection(params)

    '''
    Open a communication channel to RabbitMQ in the
    connection previously opened and stored.

    :returns: A pika.Channel object. Returns None is there was
        an expected exception.
    '''
    def __setup_rabbit_channel(self):
        try:
            return self.__make_channel()
        except pika.exceptions.ChannelClosed:
            msg = ('Problem setting up the rabbit channel.')
            self.__error_messages_during_init.append(msg)
            LOGGER.warn(msg)
            return None

    '''
    Little wrapper around pika's channel opening
    method, so we can mock this during unit tests.
    '''
    def __make_channel(self): # this is easy to mock during unit tests
        return self.__connection.channel()

    ########################
    ### Close connection ###
    ########################

    def close_rabbit_connection(self):
        if (self.__connection is not None) and (self.__connection.is_open):
            self.__connection.close()
            # From pika doc: If there are any open channels, it will attempt to
            # close them prior to fully disconnecting

    ####################
    ### Send message ###
    ####################

    '''
    Send a message to RabbitMQ, synchronously. If the
    delivery was not successful, an exception is raised.

    :param: JSON message. TODO what needs to be included?
    :raises: esgfpid.exceptions.MessageNotDeliveredException:
        In case the message was not delivered.
    '''
    def send_message_to_queue(self, message):
        self.__open_connection_if_not_open()
        routing_key, msg_string = rabbitutils.get_routing_key_and_string_message_from_message_if_possible(message)
        success = False
        error_msg = None

        try:
            success = self.__try_sending_message_several_times(routing_key, msg_string)

        except pika.exceptions.UnroutableError:
            logerror(LOGGER, 'Refused message with routing key "%s".' % routing_key)
            body_json = json.loads(msg_string)
            body_json, new_routing_key = rabbitutils.add_emergency_routing_key(body_json)
            logerror(LOGGER, 'Refused message with routing key "%s". Resending with "%s".' % (routing_key, new_routing_key))
            routing_key = new_routing_key
            try:
                success = self.__try_sending_message_several_times(routing_key, msg_string)
            except pika.exceptions.UnroutableError:
                error_msg = 'The RabbitMQ node refused a message a second time'
                error_msg += ' with the original routing key "%s" and the emergency routing key "%s").' % (body_json['original_routing_key'], routing_key)
                error_msg += ' Dropping the message.'
                logerror(LOGGER, error_msg)

        if not success:
            raise MessageNotDeliveredException(error_msg, msg_string)

    def __open_connection_if_not_open(self):
        if not self.__communication_established:
            self.open_rabbit_connection()

    def __try_sending_message_several_times(self, routing_key, message):
        delivered_after_x_times = self.__retry_x_times(routing_key, message)
        if delivered_after_x_times:
            return True
        else:
            LOGGER.error('Message could not be sent: %s', message)
            return False

    def __retry_x_times(self, routing_key, messagebody):
        message_delivered = False
        max_tries_reached = False
        counter = 1
        while (not message_delivered) and (not max_tries_reached):

            # Try to deliver:
            #LOGGER.debug('Delivery no. '+str(counter)')
            delivered_now = self.__send_message_to_queue_once(routing_key, messagebody)

            # Check if it was delivered:
            if delivered_now:
                logdebug(LOGGER, 'Successful delivery of message with routing key "%s"!', routing_key)
                logdebug(LOGGER, 'Delivered after %i times!', counter)
                message_delivered = True
            else:
                logdebug(LOGGER, 'Not delivered, waiting %i milliseconds before delivering it again!', self.__timeout_milliseconds)
                time.sleep(self.__timeout_milliseconds/1000)

            # Increment counter, check if we have reached the max:
            if counter == self.__max_tries:
                max_tries_reached = True
                logdebug(LOGGER, 'Tried %i times, giving up!', self.__max_tries)
            counter += 1

        # Return with or without success:
        return message_delivered

    def __send_message_to_queue_once(self, routing_key, messagebody):
        delivered = False
        try:
            delivered = self.__do_send_message(routing_key, messagebody, self.__props)
            self.__avoid_connection_shutdown()
            
        except pika.exceptions.UnroutableError:
            logerror(LOGGER, 'Message could not be routed to any queue, maybe none was declared yet.')
            raise

        return delivered

    def __do_send_message(self, routing_key, messagebody, props): # This is easy to mock during unit tests
        return self.__channel.basic_publish(
            exchange=self.__get_exchange_name(),
            routing_key=routing_key,
            body=messagebody,
            mandatory=self.__mandatory_flag,
            properties=props
        )
        # TODO how to catch channel close!
        # I think has to be caught in basic_publish!

        # (1) Channel closed because even fallback exchange did not exist:
        #     In asynchronous mode: reply_code == 404 and "NOT_FOUND - no exchange 'FALLBACK'" in reply_text)
        #    logerror(LOGGER,'Channel closed because FALLBACK exchange does not exist. Need to close connection to trigger all the necessary close down steps.')
        #    self.thread._connection.close() # This will reconnect!

        # (2) Channel closed because exchange did not exist:
        #     In asynchronous mode: reply_code == 404:
        #    logdebug(LOGGER, 'Channel closed because the exchange "%s" did not exist.', self.__node_manager.get_exchange_name())
        #    self.__use_different_exchange_and_reopen_channel()
        # Then we must set:
        # self.__fallback_exchange = self.__candidate_fallback_exchange_name


    def __avoid_connection_shutdown(self):
        # Source: https://github.com/pika/pika/issues/397
        if (time.time() - self.__connection_last_process_event_call) > 6:
            self.__connection.process_data_events()
            self.__connection_last_process_event_call = time.time()

    '''
    To make sure we use the fallback exchange in case it
    was set, and the normal exchange name otherwise.
    '''
    def __get_exchange_name(self):
        if self.__fallback_exchange is not None:
            return self.__fallback_exchange
        else:
            return self.__nodemanager.get_exchange_name()


import logging
import esgfpid.exceptions

LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())

#
# Utils for task 1:
#

def extract_file_handles_from_response_json(response_json, prefix):

    if response_json is None:
        raise esgfpid.exceptions.SolrResponseError('Response is None')

    handles = _extract_file_handles_from_facetfield_trackingid(response_json, prefix)
    return handles # raises esgfpid.exceptions.SolrResponseError

def _extract_file_handles_from_facetfield_trackingid(response_json, prefix):
    return _extract_handles_from_specified_facetfield(response_json, prefix, 'tracking_id')

#
# Utils for task 2:
#

def extract_dataset_handles_from_response_json(response_json, prefix):

    if response_json is None:
        raise esgfpid.exceptions.SolrResponseError('Response is None')

    handles = _extract_dataset_handles_from_facetfield_pid(response_json, prefix)
    return handles # raises esgfpid.exceptions.SolrResponseError

def _extract_dataset_handles_from_facetfield_pid(response_json, prefix):
    return _extract_handles_from_specified_facetfield(response_json, prefix, 'pid')

def extract_dataset_version_numbers_from_response_json(response_json):

    if response_json is None:
        raise esgfpid.exceptions.SolrResponseError('Response is None')

    handles = _extract_version_numbers_from_facetfield_version(response_json)
    return handles # raises esgfpid.exceptions.SolrResponseError

def _extract_version_numbers_from_facetfield_version(response_json):
    return _extract_strings_from_specified_facetfield(response_json, 'version')

def _extract_strings_from_specified_facetfield(response_json, fieldname):
    # Extract field
    list_with_counts = _extract_field_from_response_json(response_json, fieldname) # raises esgfpid.exceptions.SolrResponseError

    # Remove the counts
    list_without_counts = _remove_counts_from_list(list_with_counts)

    # Remove duplicates
    list_without_counts_nodup = _remove_duplicates_from_list(list_without_counts)

    return list_without_counts_nodup

#
# Used by both:
#

def _extract_handles_from_specified_facetfield(response_json, prefix, fieldname):
    # Extract field
    list_with_counts_without_prefixes = _extract_field_from_response_json(response_json, fieldname) # raises esgfpid.exceptions.SolrResponseError

    # Remove the counts
    list_without_counts_without_prefixes = _remove_counts_from_list(list_with_counts_without_prefixes)

    # Remove duplicates
    list_without_counts_without_prefixes_nodup = _remove_duplicates_from_list(list_without_counts_without_prefixes)

    # Add prefixes if not there:
    list_without_counts = _prepend_prefix_to_list_items_if_not_there(list_without_counts_without_prefixes_nodup, prefix)
    
    # Add "hdl:" if not there:
    list_with_hdl = _prepend_hdl_to_list_items_of_not_there(list_without_counts)

    return list_with_hdl

def _extract_field_from_response_json(response_json, fieldname):
    facet_fields = _extract_facet_fields(response_json) # raises esgfpid.exceptions.SolrResponseError
    try:
        list_with_counts = facet_fields[fieldname]
    except KeyError:
        raise esgfpid.exceptions.SolrResponseError('No field "%s" in response' % fieldname)
    return list_with_counts

def _prepend_prefix_to_list_items_if_not_there(list_without_prefixes, prefix):
    new_list_with_prefixes = []
    for suffix_of_handle in list_without_prefixes:
        if not isinstance(suffix_of_handle, int): # remove counts (if any were left)
            if prefix+'/' not in suffix_of_handle:
                new_list_with_prefixes.append(prefix+'/'+suffix_of_handle)
            else:
                new_list_with_prefixes.append(suffix_of_handle)
    return new_list_with_prefixes

def _prepend_hdl_to_list_items_of_not_there(list_without_hdl):
    new_list_with_hdl = []
    for handle in list_without_hdl:
        if not isinstance(handle, int): # remove counts (if any were left)
            if handle.startswith('hdl:'):
                new_list_with_hdl.append(handle)
            else:
                new_list_with_hdl.append('hdl:'+handle)
    return new_list_with_hdl

def _remove_counts_from_list(list_with_counts):
    new_list_without_counts = []
    for item in list_with_counts:
        if not isinstance(item, int):
            new_list_without_counts.append(item)
    return new_list_without_counts

def _remove_duplicates_from_list(list_with_dups):
    return list(set(list_with_dups))

def _extract_facet_fields(response_json):
    try:
        facet_fields = response_json['facet_counts']['facet_fields']
        return facet_fields
    except KeyError:
        raise esgfpid.exceptions.SolrResponseError('No "facet_counts" and/or "facet_fields" in response')

